import datetime
from functools import reduce
import time
from abc import abstractmethod
from typing import Callable, ClassVar, Dict, FrozenSet, List, Optional, Pattern, Union
import itertools
from dataclasses import dataclass, field

from rocketry._base import RedBase
from rocketry.pybox.time import to_datetime, to_timedelta, Interval

PARSERS: Dict[Union[str, Pattern], Union[Callable, 'TimePeriod']] = {}

@dataclass(frozen=True)
class TimePeriod(RedBase):
    """Base for all classes that represent a time period.

    Time period is a period in time with a start and an end.
    These are useful to determine whether an event took
    place in a specific time span or whether current time
    is in a given time span.
    """

    resolution: ClassVar[datetime.timedelta] = datetime.timedelta.resolution
    min: ClassVar[datetime.datetime] = datetime.datetime(1970, 1, 3, 2, 0)
    max: ClassVar[datetime.datetime] = datetime.datetime(2260, 1, 1, 0, 0)

    def __contains__(self, other):
        """Whether a given point of time is in
        the TimePeriod"""
        interval = self.rollforward(other)
        return other in interval

    def __and__(self, other):
        # self & other
        # bitwise and
        # using & operator
        is_time_period = isinstance(other, TimePeriod)
        if not is_time_period:
            raise TypeError(f"AND operator only supports TimePeriod. Given: {type(other)}")

        if self is always:
            # Reducing the operation
            return other
        elif other is always:
            # Reducing the operation
            return self
        else:
            return All(self, other)

    def __or__(self, other):
        # self | other
        # bitwise or
        is_time_period = isinstance(other, TimePeriod)
        if not is_time_period:
            raise TypeError(f"AND operator only supports TimePeriod. Given: {type(other)}")

        if self is always or other is always:
            # Reducing the operation
            return always
        else:
            return Any(self, other)

    @abstractmethod
    def rollforward(self, dt):
        "Get previous time interval of the period."
        raise NotImplementedError

    @abstractmethod
    def rollback(self, dt):
        "Get previous time interval of the period."
        raise NotImplementedError

class TimeInterval(TimePeriod):
    """Base for all time intervals

    Time interval is a period between two fixed but repeating
    times. Fixed repeated times are defined as time elements
    that repeats constantly but are fixed in a sense that
    if the datetime of an event is known, the question
    whether the event belongs to the period can be answered
    unambigiously.

    Answers to "between 11:00 and 12:00" and "from monday to tuesday"
    """
    _type_name: ClassVar = "interval"
    @abstractmethod
    def __contains__(self, dt):
        "Check whether the datetime is on the period"
        raise NotImplementedError("Contains not implemented.")

    @abstractmethod
    def rollstart(self, dt):
        "Roll forward to next point in time that on the period"
        raise NotImplementedError("Contains not implemented.")

    @abstractmethod
    def rollend(self, dt):
        "Roll back to previous point in time that is on the period"
        raise NotImplementedError("Contains not implemented.")

    @abstractmethod
    def next_start(self, dt):
        "Get next start point of the period"
        raise NotImplementedError("Contains not implemented.")

    @abstractmethod
    def next_end(self, dt):
        "Get next end point of the period"
        raise NotImplementedError("Contains not implemented.")

    @abstractmethod
    def prev_start(self, dt):
        "Get previous start point of the period"
        raise NotImplementedError("Contains not implemented.")

    @abstractmethod
    def prev_end(self, dt):
        "Get pervious end point of the period"
        raise NotImplementedError("Contains not implemented.")

    @abstractmethod
    def from_between(start, end) -> Interval:
        raise NotImplementedError("__between__ not implemented.")

    def is_full(self):
        "Whether every time belongs to the period (but there is still distinct intervals)"
        return False

    def rollforward(self, dt) -> datetime.datetime:
        "Get next time interval of the period"

        closed = "left"
        if self.is_full():
            # Full period so dt always belongs on it
            start = dt
            end = self.next_end(dt)
            if end == start:
                # Expanding the interval
                end = self.next_end(dt + self.resolution)
        else:
            start = self.rollstart(dt)
            end = self.next_end(dt)
            if start == end:
                # The interval is left closed so this should
                # not contain any points. We look for another
                # one
                return self.rollforward(end + self.resolution)

        start = to_datetime(start)
        end = to_datetime(end)

        return Interval(start, end, closed=closed)

    def rollback(self, dt) -> Interval:
        "Get previous time interval of the period"

        closed = "left"
        if self.is_full():
            # Full period so dt always belongs on it
            end = dt
            start = self.prev_start(dt)
            if end == start:
                # Expanding the interval
                start = self.prev_start(dt - self.resolution)
        else:
            end = self.rollend(dt)
            start = self.prev_start(dt)
            if start == end:
                # The interval is left closed but the start
                # is included in the interval. Therefore
                # we include a single point (both sides closed)
                closed = "both"

        start = to_datetime(start)
        end = to_datetime(end)

        return Interval(start, end, closed=closed)

    def __eq__(self, other):
        "Test whether self and other are essentially the same periods"
        is_same_class = type(self) == type(other)
        if is_same_class:
            return (self._start == other._start) and (self._end == other._end)
        else:
            return False

@dataclass(frozen=True)
class TimeDelta(TimePeriod):
    """Base for all time deltas

    Time delta is a period of time from a reference
    point. It is floating in nature as it could
    contain arbitrary datetimes depending on where
    the reference point is set. This reference
    point is typically current datetime.
    """
    _type_name: ClassVar = "delta"

    reference: Optional[datetime.datetime] = field(default=None)

    def __init__(self, past=None, future=None, reference=None, *, kws_past=None, kws_future=None):

        past = 0 if past is None else past
        future = 0 if future is None else future

        kws_past = {} if kws_past is None else kws_past
        kws_future = {} if kws_future is None else kws_future

        object.__setattr__(self, "past", abs(to_timedelta(past, **kws_past)))
        object.__setattr__(self, "future", abs(to_timedelta(future, **kws_future)))
        object.__setattr__(self, "reference", reference)

    @abstractmethod
    def __contains__(self, dt):
        "Check whether the datetime is in "
        reference = self.get_reference()
        start = reference - abs(self.past)
        end = reference + abs(self.future)
        return start <= dt <= end

    def get_reference(self) -> datetime.datetime:
        return self.reference if self.reference is not None else datetime.datetime.fromtimestamp(time.time())

    def rollback(self, dt):
        "Get previous interval (including currently ongoing)"
        start = dt - abs(self.past)
        start = to_datetime(start)
        end = to_datetime(dt)
        return Interval(start, end)

    def rollforward(self, dt):
        "Get next interval (including currently ongoing)"
        end = dt + abs(self.future)
        start = to_datetime(dt)
        end = to_datetime(end)
        return Interval(start, end)

    def __eq__(self, other):
        "Test whether self and other are essentially the same periods"
        is_same_class = type(self) == type(other)
        if is_same_class:
            return (self.past == other.past) and (self.future == other.future)
        else:
            return False

    def __str__(self):
        if not self.future:
            return f"past {str(self.past)}"
        elif not self.past:
            return f"next {str(self.future)}"
        else:
            return f"past {str(self.past)} to next {str(self.future)}"

    def __repr__(self):
        return f"TimeDelta(past={repr(self.past)}, future={repr(self.future)})"

def all_overlap(times:List[Interval]):
    return all(a.overlaps(b) for a, b in itertools.combinations(times, 2))

def get_overlapping(times):
    # Example:
    # A:    <-------------->
    # B:     <------>
    # C:         <------>
    # Out:       <-->
    starts = [interval.left for interval in times]
    ends = [interval.right for interval in times]

    start = max(starts)
    end = min(ends)
    return Interval(start, end)

@dataclass(frozen=True)
class All(TimePeriod):

    periods: FrozenSet[TimePeriod]

    def __init__(self, *args):
        if any(not isinstance(arg, TimePeriod) for arg in args):
            raise TypeError("Only TimePeriods supported")
        elif not args:
            raise ValueError("No TimePeriods to wrap")

        # Compress the time periods
        periods = []
        for arg in args:
            if isinstance(arg, All):
                # Don't nest unnecessarily
                periods += list(arg.periods)
            elif arg is always:
                # Does not really have an effect
                continue
            else:
                periods.append(arg)

        object.__setattr__(self, "periods", frozenset(periods))

    def rollback(self, dt):

        # We solve this iteratively
        # 1. rollback
        # 2. check if everything overlaps
        # 3. If not overlaps, take max and check again
        # 4. If overlaps, get the period that overlaps

        intervals = [
            period.rollback(dt)
            for period in self.periods
        ]
        all_overlaps = all(a.overlaps(b) for a, b in itertools.combinations(intervals, 2))
        if all_overlaps:
            return reduce(lambda a, b: a & b, intervals)
        else:
            # Not found, trying again with next period
            # Example:
            # Current:                     |
            # A:         <-------------->
            # B:         <---> <--->
            # C:         <------>
            # Next try:         |
            next_dt = min(intervals, key=lambda x: x.right).right

            opened = any(
                interv.closed not in ('right', 'both')
                for interv in intervals
                if interv.right == next_dt
            )
            # TODO: If
            if dt == next_dt:
                next_dt -= self.resolution
            return self.rollback(next_dt)

    def rollforward(self, dt):
        # We solve this iteratively
        # 1. rollforward
        # 2. check if everything overlaps
        # 3. If not overlaps, take max and check again
        # 4. If overlaps, get the period that overlaps

        intervals = [
            period.rollforward(dt)
            for period in self.periods
        ]
        all_overlaps = all(a.overlaps(b) for a, b in itertools.combinations(intervals, 2))
        if all_overlaps:
            return reduce(lambda a, b: a & b, intervals)
        else:
            # Not found, trying again with next period
            # Example:
            # Current: |
            # A:         <-------------->
            # B:         <---> <--->
            # C:                 <------>
            # Next try:          |
            next_dt = max(intervals, key=lambda x: x.left).left
            opened = any(
                interv.closed not in ('left', 'both')
                for interv in intervals
                if interv.left == next_dt
            )
            if opened:
                next_dt -= self.resolution
            return self.rollforward(next_dt)

    def __eq__(self, other):
        # self | other
        # bitwise or
        if type(self) == type(other):
            return self.periods == other.periods
        else:
            return False

    def __repr__(self):
        sub = ', '.join(repr(p) for p in self.periods)
        return f"All({sub})"

    def __str__(self):
        return ' & '.join(str(p) for p in self.periods)

@dataclass(frozen=True)
class Any(TimePeriod):

    periods: FrozenSet[TimePeriod]

    def __init__(self, *args):
        if any(not isinstance(arg, TimePeriod) for arg in args):
            raise TypeError("Only TimePeriods supported")
        elif not args:
            raise ValueError("No TimePeriods to wrap")

        # Compress the time periods
        periods = []
        for arg in args:
            if isinstance(arg, Any):
                # Don't nest unnecessarily
                periods += list(arg.periods)
            elif arg is always:
                # Does not really have an effect
                periods = [always]
                break
            else:
                periods.append(arg)

        object.__setattr__(self, "periods", frozenset(periods))

    def rollback(self, dt):
        intervals = [
            period.rollback(dt)
            for period in self.periods
        ]

        # Example:
        # Current time           |
        # A:    <-------------->
        # B:     <------>
        # C:         <------------->
        # Out:  <------------------>

        # Example:
        # Current time           |
        # A:    <-->
        # B:     <--->     <--->
        # C:     <------>
        # Out:             <--->

        # Example:
        # Current time           |
        # A:    <-->
        # B:    <--->     <--->
        # C:        <----->
        # Out:  <------------->

        # We solve the problem iteratively
        # 1. Find the interval that ends closest to the dt
        # 2. Check if there is an interval overlapping with longer end
        # 3. Repeat 2 until there is none

        # Sorting the closest first (right is oldest)
        intervals = sorted(intervals, key=lambda x: x.right, reverse=True)

        curr_interval = intervals.pop(0)
        end_interval = curr_interval

        for interv in intervals:
            extends = interv.left < curr_interval.left
            if extends and interv.overlaps(curr_interval):
                curr_interval = interv
            else:
                break

        return Interval(
            curr_interval.left,
            end_interval.right
        )

    def rollforward(self, dt):
        intervals = [
            period.rollforward(dt)
            for period in self.periods
        ]

        # We solve the problem iteratively
        # 1. Find the interval that starts closest to the dt
        # 2. Check if there is an interval overlapping with longer end
        # 3. Repeat 2 until there is none

        # Sorting the closest first (left is newest)
        intervals = sorted(intervals, key=lambda x: x.left, reverse=False)

        curr_interval = intervals.pop(0)
        start_interval = curr_interval

        for interv in intervals:
            extends = interv.right > curr_interval.right
            if extends and interv.overlaps(curr_interval):
                curr_interval = interv
            else:
                break

        return Interval(
            start_interval.left,
            curr_interval.right
        )

    def __eq__(self, other):
        # self | other
        # bitwise or
        if type(self) == type(other):
            return self.periods == other.periods
        else:
            return False

    def __repr__(self):
        sub = ', '.join(repr(p) for p in self.periods)
        return f"Any({sub})"

    def __str__(self):
        return ' | '.join(str(p) for p in self.periods)

@dataclass(frozen=True)
class StaticInterval(TimePeriod):
    """Inverval that is fixed in specific datetimes."""

    start: datetime.datetime
    end: datetime.datetime

    def __init__(self, start=None, end=None):
        object.__setattr__(self, "start", to_datetime(start) if start is not None else self.min)
        object.__setattr__(self, "end", to_datetime(end) if end is not None else self.max)

    def rollback(self, dt):
        dt = to_datetime(dt)
        start = to_datetime(self.start)
        if start > dt:
            # The actual interval is in the future
            return Interval(self.min, self.min)
        end = min(self.end, dt)
        return Interval(start, end)

    def rollforward(self, dt):
        dt = to_datetime(dt)
        end = to_datetime(self.end)
        if end < dt:
            # The actual interval is already gone
            return Interval(self.max, self.max)
        start = max(self.start, dt)
        return Interval(start, end)

    @property
    def is_max_interval(self):
        return (self.start == self.min) and (self.end == self.max)

    def __str__(self):
        if self.is_max_interval:
            return 'always'
        elif self.start == self.max:
            return 'never'
        else:
            return f'between {self.start} and {self.end}'

    def __repr__(self):
        if self.is_max_interval:
            return 'always'
        elif self.start == self.max:
            return 'never'
        else:
            return f"StaticInterval(start={self.start!r}, end={self.end!r})"

always = StaticInterval()
never = StaticInterval(StaticInterval.max, StaticInterval.max)



from typing import Callable, Dict, Pattern, Union
from rocketry.core.condition.base import BaseCondition
from rocketry.core.condition.base import BaseCondition
from rocketry.session import Session
from ..utils import ParserError, CondParser



CONDITION_PARSERS = []

def add_condition_parser(d: Dict[Union[str, Pattern], Union[Callable, 'BaseCondition']]):
    """Add a parsing instruction to be used for parsing a
    string to condition.

    Parameters
    ----------
    d : dict
        TODO
    """
    parsers = Session._cls_cond_parsers #! TODO
    parsers.update(d)

def parse_condition_item(s:str, session=None) -> BaseCondition:
    "Parse one condition"

    # TODO: Don't use global
    session = Session.session if session is None else session

    for statement, parser in session.get_cond_parsers().items():
        if isinstance(statement, Pattern):
            res = statement.fullmatch(s)
            if res:
                args = ()
                kwargs = res.groupdict()
                break
        else:
            if s == statement:
                args = (s,)
                kwargs = {}
                break
    else:
        raise ParserError(f"Could not find parser for string {repr(s)}.")

    if isinstance(parser, BaseCondition):
        return parser
    else:
        if isinstance(parser, CondParser):
            return parser(s, **kwargs)
        cond = parser(**kwargs)
        return cond



import pytest

from rocketry.time import Any
from rocketry.time import TimeOfDay, TimeOfHour, TimeOfWeek
from rocketry.time.interval import TimeOfMinute, TimeOfMonth, TimeOfYear

def test_time_of_day_every_second():
    every_second_hour = TimeOfDay.create_range(step=2)
    assert isinstance(every_second_hour, Any)
    assert len(every_second_hour.periods) == 12
    assert every_second_hour == Any(
        TimeOfDay("00:00", "01:00"),
        #TimeOfDay("01:00", "02:00"),
        TimeOfDay("02:00", "03:00"),
        #TimeOfDay("03:00", "04:00"),
        TimeOfDay("04:00", "05:00"),
        #TimeOfDay("05:00", "06:00"),
        TimeOfDay("06:00", "07:00"),
        #TimeOfDay("07:00", "08:00"),
        TimeOfDay("08:00", "09:00"),
        #TimeOfDay("09:00", "10:00"),
        TimeOfDay("10:00", "11:00"),
        #TimeOfDay("11:00", "12:00"),
        TimeOfDay("12:00", "13:00"),
        #TimeOfDay("13:00", "14:00"),
        TimeOfDay("14:00", "15:00"),
        #TimeOfDay("15:00", "16:00"),
        TimeOfDay("16:00", "17:00"),
        #TimeOfDay("17:00", "18:00"),
        TimeOfDay("18:00", "19:00"),
        #TimeOfDay("19:00", "20:00"),
        TimeOfDay("20:00", "21:00"),
        #TimeOfDay("21:00", "22:00"),
        TimeOfDay("22:00", "23:00"),
        #TimeOfDay("23:00", "24:00"),
    )

def test_time_of_day_every_third():
    every_second_hour = TimeOfDay.create_range(step=3)
    assert isinstance(every_second_hour, Any)
    assert every_second_hour == Any(
        TimeOfDay("00:00", "01:00"),
        #TimeOfDay("01:00", "02:00"),
        #TimeOfDay("02:00", "03:00"),
        TimeOfDay("03:00", "04:00"),
        #TimeOfDay("04:00", "05:00"),
        #TimeOfDay("05:00", "06:00"),
        TimeOfDay("06:00", "07:00"),
        #TimeOfDay("07:00", "08:00"),
        #TimeOfDay("08:00", "09:00"),
        TimeOfDay("09:00", "10:00"),
        #TimeOfDay("10:00", "11:00"),
        #TimeOfDay("11:00", "12:00"),
        TimeOfDay("12:00", "13:00"),
        #TimeOfDay("13:00", "14:00"),
        #TimeOfDay("14:00", "15:00"),
        TimeOfDay("15:00", "16:00"),
        #TimeOfDay("16:00", "17:00"),
        #TimeOfDay("17:00", "18:00"),
        TimeOfDay("18:00", "19:00"),
        #TimeOfDay("19:00", "20:00"),
        #TimeOfDay("20:00", "21:00"),
        TimeOfDay("21:00", "22:00"),
        #TimeOfDay("22:00", "23:00"),
        #TimeOfDay("23:00", "24:00"),
    )

@pytest.mark.parametrize("cls,step,n_periods",
    [
        pytest.param(TimeOfMinute, 1, 60, id="TimeOfMinute (every second)"),
        pytest.param(TimeOfHour, 1, 60, id="TimeOfHour (every minute)"),
        pytest.param(TimeOfDay, 1, 24, id="TimeOfDay (every hour)"),
        pytest.param(TimeOfWeek, 1, 7, id="TimeOfWeek (every day of week)"),
        pytest.param(TimeOfMonth, 1, 31, id="TimeOfMonth (every day of month)"),
        pytest.param(TimeOfYear, 1, 12, id="TimeOfYear (every month)"),

        pytest.param(TimeOfMinute, 2, 30, id="TimeOfMinute (every second second)"),
        pytest.param(TimeOfHour, 2, 30, id="TimeOfHour (every second minute)"),
        pytest.param(TimeOfDay, 2, 12, id="TimeOfDay (every second hour)"),
        pytest.param(TimeOfWeek, 2, 4, id="TimeOfWeek (every second day of week)"),
        pytest.param(TimeOfMonth, 2, 16, id="TimeOfMonth (every second day of month)"),
        pytest.param(TimeOfYear, 2, 6, id="TimeOfYear (every second month)"),

        pytest.param(TimeOfMinute, 3, 20, id="TimeOfMinute (every third second)"),
        pytest.param(TimeOfHour, 3, 20, id="TimeOfHour (every third minute)"),
        pytest.param(TimeOfDay, 3, 8, id="TimeOfDay (every third hour)"),
        pytest.param(TimeOfWeek, 3, 3, id="TimeOfWeek (every third day of week)"),
        pytest.param(TimeOfMonth, 3, 11, id="TimeOfMonth (every third day of month)"),
        pytest.param(TimeOfYear, 3, 4, id="TimeOfYear (every third month)"),

        pytest.param(TimeOfMinute, 61, 1, id="TimeOfMinute (over the period)"),
        pytest.param(TimeOfHour, 61, 1, id="TimeOfHour (over the period)"),
        pytest.param(TimeOfDay, 24, 1, id="TimeOfDay (over the period)"),
        pytest.param(TimeOfWeek, 7, 1, id="TimeOfWeek (over the period)"),
        pytest.param(TimeOfMonth, 31, 1, id="TimeOfMonth (over the period)"),
        pytest.param(TimeOfYear, 12, 1, id="TimeOfYear (over the period)"),
    ]
)
def test_every_second(cls, step, n_periods):
    every = cls.create_range(step=step)
    assert isinstance(every, Any)
    assert len(every.periods) == n_periods



import datetime
import logging
import pickle
from textwrap import dedent
import pytest
from rocketry.args.builtin import Return
from rocketry.core import Task as BaseTask
from rocketry.core.condition.base import AlwaysFalse
from rocketry.args import Arg, Session, Task
from rocketry.exc import TaskLoggingError
from rocketry.log import MinimalRecord
from rocketry import Session as SessionClass
from rocketry.testing.log import create_task_record

class DummyTask(BaseTask):

    def execute(self, *args, **kwargs):
        return

def test_defaults(session):
    task = DummyTask(name="mytest", session=session)
    assert task.name == "mytest"
    assert isinstance(task.start_cond, AlwaysFalse)
    assert isinstance(task.end_cond, AlwaysFalse)

def test_defaults_no_session(session):
    with pytest.warns(UserWarning):
        task = DummyTask(name="mytest")
    assert task.session is not session
    assert isinstance(task.session, SessionClass)
    assert task.session.tasks == {task}

def test_set_timeout(session):
    task = DummyTask(timeout="1 hour 20 min", session=session, name="1")
    assert task.timeout == datetime.timedelta(hours=1, minutes=20)

    task = DummyTask(timeout=datetime.timedelta(hours=1, minutes=20), session=session, name="2")
    assert task.timeout == datetime.timedelta(hours=1, minutes=20)

    task = DummyTask(timeout=20, session=session, name="3")
    assert task.timeout == datetime.timedelta(seconds=20)

def test_delete(session):
    task = DummyTask(name="mytest", session=session)
    assert session.tasks == {task}
    task.delete()
    assert session.tasks == set()

def test_set_invalid_status(session):
    task = DummyTask(name="mytest", session=session)
    with pytest.raises(ValueError):
        task.status = "not valid"

def test_failed_logging(session):

    class MyHandler(logging.Handler):
        def emit(self, record):
            raise RuntimeError("Oops")

    logging.getLogger("rocketry.task").handlers.insert(0, MyHandler())
    task = DummyTask(name="mytest", session=session)
    for func in (task.log_crash, task.log_failure, task.log_success, task.log_inaction, task.log_termination):
        with pytest.raises(TaskLoggingError):
            func()

    record = create_task_record(created=1, task_name="mytest", action="run")

    with pytest.raises(TaskLoggingError):
        task.log_record(record) # Used by process logging

def test_pickle(session):
    task_1 = DummyTask(name="mytest", session=session)
    pkl_obj = pickle.dumps(task_1)
    task_2 = pickle.loads(pkl_obj)
    assert task_1.name == task_2.name

def test_crash(session):
    task = DummyTask(name="mytest", session=session)
    task.log_running()
    assert task.status == "run"
    assert task.last_crash is None
    task.delete()

    # Recreating and now should log crash
    task = DummyTask(name="mytest", session=session)
    assert task.status == "crash"
    assert task.last_crash

    logs = task.logger.filter_by().all()
    assert [
        {'action': 'run', 'task_name': 'mytest'},
        {'action': 'crash', 'task_name': 'mytest'}
    ] == [log.dict(exclude={'created'}) for log in logs]

def test_json(session):
    session.parameters['x'] = 5
    repo = session.get_repo()
    repo.add(MinimalRecord(task_name="mytest", action="run", created=1640988000))
    repo.add(MinimalRecord(task_name="mytest", action="success", created=1640988060))

    task = DummyTask(name="mytest", parameters={
        "arg_2": Arg("x"),
        "arg_2": Return("another"),
        "session": Session(),
        "task": Task(),
        "another_task": Task('another')
    }, session=session)
    j = task.json(indent=4)

    dt_run = datetime.datetime.fromtimestamp(1640988000)
    dt_success = datetime.datetime.fromtimestamp(1640988060)

    assert j == dedent("""
    {
        "permanent_task": false,
        "fmt_log_message": "Task '{task}' status: '{action}'",
        "daemon": null,
        "batches": [],
        "name": "mytest",
        "description": null,
        "logger_name": "rocketry.task",
        "execution": null,
        "priority": 0,
        "disabled": false,
        "force_run": false,
        "force_termination": false,
        "status": "success",
        "timeout": null,
        "parameters": {
            "arg_2": "Return('another', default=NOTSET)",
            "session": "session",
            "task": "Task()",
            "another_task": "Task('another')"
        },
        "start_cond": "false",
        "end_cond": "false",
        "on_startup": false,
        "on_shutdown": false,
        "last_run": "<RUN>",
        "last_success": "<SUCCESS>",
        "last_fail": null,
        "last_terminate": null,
        "last_inaction": null,
        "last_crash": null
    }
    """
    .replace("<RUN>", dt_run.strftime("%Y-%m-%dT%H:%M:%S"))
    .replace("<SUCCESS>", dt_success.strftime("%Y-%m-%dT%H:%M:%S"))
    )[1:-1]



import pytest

from rocketry.conditions import AlwaysFalse
from rocketry.tasks.maintain import ShutDown
from rocketry.tasks import FuncTask
from rocketry.exc import SchedulerExit

def write_file(text):
    with open("test.txt", "a") as f:
        f.write(text)

def test_restart_raises(session):
    task = ShutDown(session=session)
    with pytest.raises(SchedulerExit):
        task()

def test_scheduler_shutdown(tmpdir, session):

    with tmpdir.as_cwd() as old_dir:

        FuncTask(write_file, parameters={"text": "Started"}, on_startup=True, name="write_startup", execution="main", session=session)
        FuncTask(write_file, parameters={"text": "Shut"}, on_shutdown=True, name="write_shutdown", execution="main", session=session)

        task = ShutDown(session=session)

        task.run()

        session.config.shut_cond = AlwaysFalse()

        session.start()

        with open("test.txt") as f:
            cont = f.read()
        assert "StartedShut" == cont

        records = list(map(lambda e: e.dict(exclude={'created'}), task.logger.get_records()))
        assert 1 == len([record for record in records if record["action"] == "run"])
        assert 1 == len([record for record in records if record["action"] == "success"])




import logging

import pytest
from rocketry.core.log.adapter import TaskAdapter
from rocketry import Session
from rocketry.tasks import FuncTask
from rocketry.core import Parameters


def test_tasks_attr(session):

    task1 = FuncTask(
        lambda : None,
        name="example 1",
        execution="main",
        session=session
    )
    task2 = FuncTask(
        lambda : None,
        name="example 2",
        execution="main",
        session=session
    )

    assert session.tasks == {task1, task2}

def test_get_repo(session):

    logger = logging.getLogger("rocketry.task")
    assert session.get_repo() is logger.handlers[0].repo

    # Test the one used in the task logging is also the same
    assert session.get_repo() is TaskAdapter(logger, task=None)._get_repo()

# Manipulate
# ----------

def test_getitem(session):

    task_1 = FuncTask(
        lambda : None,
        name="task 1",
        execution="main",
        session=session
    )
    task_2 = FuncTask(
        lambda : None,
        name="task 2",
        execution="main",
        session=session
    )

    @FuncTask(execution="main", session=session)
    def do_things():
        pass

    @FuncTask(name="task 3", execution="main", session=session)
    def do_things_2():
        pass

    assert session['task 1'] is task_1
    assert session[task_2] is task_2
    assert isinstance(session[do_things], FuncTask)
    assert session[do_things_2].name == "task 3"

    with pytest.raises(TypeError):
        session[1234]
    with pytest.raises(KeyError):
        session["non existing"]

def test_add(session):
    task = FuncTask(
        lambda : None,
        name="task 1",
        execution="main",
        session=Session(config=dict(task_execution="async")),
    )

    assert session.tasks == set()
    session.add_task(task)

    assert session.tasks == {task}

def test_create(session):

    assert session.tasks == set()
    session.create_task(name="do_command", start_cond="daily", command="echo 'hello world'")

    @session.create_task(name="do_func", start_cond="daily")
    def do_things():
        ...
    assert session.tasks == {session['do_func'], session['do_command']}

def test_remove(session):
    task_1 = FuncTask(
        lambda : None,
        name="task 1",
        execution="main",
        session=session
    )
    task_2 = FuncTask(
        lambda : None,
        name="task 2",
        execution="main",
        session=session
    )
    task_3 = FuncTask(
        lambda : None,
        name="task 3",
        execution="main",
        session=session
    )
    assert session.tasks == {task_1, task_2, task_3}
    session.remove_task(task_3)
    assert session.tasks == {task_1, task_2}
    session.remove_task("task 2")
    assert session.tasks == {task_1}

# Old interface
# -------------

def test_task_exists(session):
    task_1 = FuncTask(
        lambda : None,
        name="task 1",
        execution="main",
        session=session
    )
    with pytest.warns(DeprecationWarning):
        assert session.task_exists("task 1")
        assert not session.task_exists("task not exists")
        assert session.task_exists(task_1)

def test_get_task(session):

    task = FuncTask(
        lambda : None,
        name="example",
        execution="main",
        session=session
    )

    with pytest.warns(DeprecationWarning):
        # By string
        t = session.get_task(task.name)
        assert t is task

    with pytest.warns(DeprecationWarning):
        # By task (returns itself)
        t = session.get_task(task)
        assert t is task

def test_clear(session):

    assert session.tasks == set()
    assert Parameters() == session.parameters
    # assert session.scheduler is None

    task1 = FuncTask(
        lambda : None,
        name="example 1",
        execution="main",
        session=session
    )
    task2 = FuncTask(
        lambda : None,
        name="example 2",
        execution="main",
        session=session
    )
    session.parameters["x"] = 1

    assert Parameters(x=1) == session.parameters
    assert session.tasks == {task1, task2}

    session.clear()

    assert session.tasks == set()
    assert Parameters() == session.parameters



import datetime

import pytest
from rocketry.testing.log import create_task_record

def test_create_record(session):
    ts = int(datetime.datetime(2022, 1, 1).timestamp())
    rec = create_task_record(task_name="mytask", action="run", created="2022-01-01")
    assert rec.task_name == "mytask"
    assert rec.action == "run"
    assert rec.created == ts
    assert rec.levelname == "INFO"
    assert rec.name == session.config.task_logger_basename

    rec = create_task_record(task_name="mytask", action="success", created=datetime.datetime(2022, 1, 1))
    assert rec.task_name == "mytask"
    assert rec.action == "success"
    assert rec.created == ts
    assert rec.levelname == "INFO"

    rec = create_task_record(task_name="mytask", action="fail", created=ts)
    assert rec.task_name == "mytask"
    assert rec.action == "fail"
    assert rec.created == ts
    assert rec.levelname == "ERROR"

    with pytest.raises(ValueError):
        create_task_record(task_name="mytask", action="typo", created=ts)




from typing import Pattern
from rocketry.core.time.base import TimePeriod
from rocketry.session import Session
from ..utils import ParserError


def parse_time_item(s:str, session=None):
    "Parse one condition"
    if session is None:
        # Old way
        session = Session.session
    parsers = session._time_parsers
    for statement, parser in parsers.items():
        if isinstance(statement, Pattern):
            res = statement.fullmatch(s)
            if res:
                args = ()
                kwargs = res.groupdict()
                break
        else:
            if s == statement:
                args = (s,)
                kwargs = {}
                break
    else:
        raise ParserError(f"Could not find parser for string {repr(s)}.")

    if isinstance(parser, TimePeriod):
        return parser
    else:
        return parser(**kwargs)



from functools import partial

from rocketry.time.interval import TimeOfHour, TimeOfMinute, TimeOfMonth
from rocketry.core.condition import AlwaysFalse, AlwaysTrue, All, Any, Not, BaseCondition
from .func import FuncCond
from .task import *
from .scheduler import *
from .time import *
from .parameter import ParamExists, IsEnv
from .meta import TaskCond



from dataclasses import dataclass
from typing import Any

try:
    from typing import Literal
except ImportError: # pragma: no cover
    from typing_extensions import Literal

@dataclass(frozen=True)
class Interval:
    "Mimics pandas.Interval"

    left: Any
    right: Any
    closed: Literal['left', 'right', 'both', 'neither'] = "left"

    def __post_init__(self):
        if self.left > self.right:
            raise ValueError("Left cannot be greater than right")

        if self.closed not in ('left', 'right', 'both', 'neither'):
            raise ValueError(f"Invalid close: {self.closed}")


    def __contains__(self, dt):
        if self.closed == "right":
            return self.left < dt <= self.right
        elif self.closed == "left":
            return self.left <= dt < self.right
        elif self.closed == "both":
            return self.left <= dt <= self.right
        else:
            return self.left < dt < self.right

    def overlaps(self, other:'Interval'):
        # Simple case: No overlap if:

        # self:  |------|
        # other:          |------|
        is_self_on_far_left = self.right < other.left

        # Or:
        # self:           |------|
        # other: |------|
        is_self_on_far_right = self.left > other.right
        is_not_overlap = is_self_on_far_left or is_self_on_far_right
        if is_not_overlap:
            return False

        # More complex:
        # self:  |------?
        # other:        ?------|
        if self.right == other.left:
            if self.closed in ('left', 'neither') or other.closed in ('right', 'neither'):
                # self:  |------*
                # other:        |------|

                # self:  |------|
                # other:        *------|
                return False

        # self:         ?------|
        # other: |------?
        if self.left == other.right:
            if self.closed in ('right', 'neither') or other.closed in ('left', 'neither'):
                # self:         *------|
                # other: |------|

                # self:         |------|
                # other: |------*
                return False

        return True

    def __and__(self, other):
        "Find interval that both overlaps"

        if self.left > other.left:
            left = self.left
            left_closed = self.closed in ('left', 'both')
        elif self.left < other.left:
            left = other.left
            left_closed = other.closed in ('left', 'both')
        else:
            # Equal
            left = self.left
            left_closed = self.closed in ('left', 'both') and other.closed in ('left', 'both')

        if self.right < other.right:
            right = self.right
            right_closed = self.closed in ('right', 'both')
        elif self.right > other.right:
            right = other.right
            right_closed = other.closed in ('right', 'both')
        else:
            # Equal
            right = self.right
            right_closed = self.closed in ('right', 'both') and other.closed in ('right', 'both')

        closed = (
            "both" if left_closed and right_closed
            else 'left' if left_closed
            else 'right' if right_closed
            else 'neither'
        )

        return Interval(left, right, closed=closed)

    @property
    def is_empty(self):
        "Check if constains no points"
        if self.closed == "both":
            has_points = self.left <= self.right
        elif self.closed in ('left', 'right', 'neither'):
            has_points = self.left < self.right

        return not has_points

    def __repr__(self):
        return f'Interval({repr(self.left)}, {repr(self.right)}, closed={repr(self.closed)})'



import pytest
from rocketry.conditions.scheduler import SchedulerCycles, SchedulerStarted
from rocketry.conditions.task.task import TaskFailed, TaskFinished, TaskRunning, TaskStarted, TaskSucceeded

from rocketry.conds import (
    true, false,
    every,
    secondly, minutely, hourly, daily, weekly, monthly,
    time_of_second, time_of_minute, time_of_hour, time_of_day, time_of_week, time_of_month,
    after_finish, after_success, after_fail,

    after_all_success, after_any_success, after_all_fail, after_any_fail, after_all_finish, after_any_finish,

    scheduler_running, scheduler_cycles,

    succeeded, failed, finished, started,
    running,

    cron,
    retry,
    crontime,
)

from rocketry.conditions import TaskExecutable, IsPeriod, DependSuccess, DependFailure, DependFinish, TaskRunnable, Retry
from rocketry.core.condition import AlwaysFalse, AlwaysTrue, Any, All
from rocketry.time import TimeDelta
from rocketry.time import Cron
from rocketry.time.delta import TimeSpanDelta
from rocketry.tasks import FuncTask
from rocketry.time.interval import TimeOfDay, TimeOfHour, TimeOfMinute, TimeOfMonth, TimeOfSecond, TimeOfWeek

params_basic = [
    pytest.param(true, AlwaysTrue(), id="true"),
    pytest.param(false, AlwaysFalse(), id="false"),
]

params_time = [
    pytest.param(time_of_second.starting(500), IsPeriod(period=TimeOfSecond.starting(500)), id="time of second starting"),
    pytest.param(time_of_minute.starting(45.00), IsPeriod(period=TimeOfMinute.starting(45.00)), id="time of minute starting"),
    pytest.param(time_of_hour.after("45:00"), IsPeriod(period=TimeOfHour("45:00", None)), id="time of hour after"),

    pytest.param(time_of_day.after("10:00"), IsPeriod(period=TimeOfDay("10:00", None)), id="time of day after"),
    pytest.param(time_of_day.before("10:00"), IsPeriod(period=TimeOfDay(None, "10:00")), id="time of day before"),
    pytest.param(time_of_day.between("10:00", "12:00"), IsPeriod(period=TimeOfDay("10:00", "12:00")), id="time of day between"),

    pytest.param(time_of_week.on("Monday"), IsPeriod(period=TimeOfWeek("Mon", time_point=True)), id="time of week on"),
    pytest.param(time_of_month.between("1st", "5th"), IsPeriod(period=TimeOfMonth("1st", "5th")), id="time of month between"),
]

params_task_exec = [
    pytest.param(every("10 mins"), TaskStarted(period=TimeDelta("10 mins")) == 0, id="every"),
    pytest.param(every("10 mins", based="finish"), TaskExecutable(period=TimeDelta("10 mins")), id="every (finish)"),
    pytest.param(every("10 mins", based="success"), TaskSucceeded(period=TimeDelta("10 mins")) == 0, id="every (success)"),
    pytest.param(every("10 mins", based="fail"), TaskFailed(period=TimeDelta("10 mins")) == 0, id="every (fail)"),

    pytest.param(secondly.get_cond(), TaskExecutable(period=TimeOfSecond(None, None)), id="secondly"),
    pytest.param(minutely.get_cond(), TaskExecutable(period=TimeOfMinute(None, None)), id="minutely"),
    pytest.param(hourly.get_cond(), TaskExecutable(period=TimeOfHour(None, None)), id="hourly"),
    pytest.param(daily.get_cond(), TaskExecutable(period=TimeOfDay(None, None)), id="daily"),
    pytest.param(weekly.get_cond(), TaskExecutable(period=TimeOfWeek(None, None)), id="weekly"),

    pytest.param(minutely.after("45:00"), TaskExecutable(period=TimeOfMinute("45:00", None)), id="minutely after"),
    pytest.param(minutely.before("45:00"), TaskExecutable(period=TimeOfMinute(None, "45:00")), id="minutely before"),

    pytest.param(hourly.after("10:00"), TaskExecutable(period=TimeOfHour("10:00", None)), id="hourly after"),
    pytest.param(hourly.before("10:00"), TaskExecutable(period=TimeOfHour(None, "10:00")), id="hourly before"),
    pytest.param(hourly.between("10:00", "12:00"), TaskExecutable(period=TimeOfHour("10:00", "12:00")), id="hourly between"),
    pytest.param(hourly.starting("10:00"), TaskExecutable(period=TimeOfHour("10:00", "10:00")), id="hourly starting"),

    pytest.param(daily.between("10:00", "12:00"), TaskExecutable(period=TimeOfDay("10:00", "12:00")), id="daily between"),

    pytest.param(secondly.at("45"), TaskExecutable(period=TimeOfSecond("45", "46")), id="minutely at"),
    pytest.param(minutely.at("45"), TaskExecutable(period=TimeOfMinute("45", "46")), id="minutely at"),
    pytest.param(hourly.at("30:00.0000"), TaskExecutable(period=TimeOfHour("30:00", "31:00")), id="hourly at"),
    pytest.param(daily.at("10:00:00"), TaskExecutable(period=TimeOfDay("10:00", "11:00")), id="daily at"),
    pytest.param(weekly.at("Monday"), TaskExecutable(period=TimeOfWeek("Monday", time_point=True)), id="weekly at"),
    pytest.param(monthly.at("1st"), TaskExecutable(period=TimeOfMonth("1st", time_point=True)), id="monthly at"),

    pytest.param(weekly.on("Monday"), TaskExecutable(period=TimeOfWeek("Monday", time_point=True)), id="weekly on"),
    pytest.param(monthly.on("1st"), TaskExecutable(period=TimeOfMonth("1st", time_point=True)), id="monthly on"),

    pytest.param(daily.at("23:00:00"), TaskExecutable(period=TimeOfDay("23:00:00", "00:00:00")), id="daily at end"),
    pytest.param(daily.at("23:00:01"), TaskExecutable(period=TimeOfDay("23:00:01", "00:00:01")), id="daily at specific"),
]

params_pipeline = [
    pytest.param(after_finish("other"), DependFinish(depend_task="other"), id="after finish"),
    pytest.param(after_success("other"), DependSuccess(depend_task="other"), id="after success"),
    pytest.param(after_fail("other"), DependFailure(depend_task="other"), id="after fail"),

    pytest.param(after_all_finish("do_a", "do_b"), All(DependFinish(depend_task="do_a"), DependFinish(depend_task="do_b")), id="after all finish"),
    pytest.param(after_all_success("do_a", "do_b"), All(DependSuccess(depend_task="do_a"), DependSuccess(depend_task="do_b")), id="after all success"),
    pytest.param(after_all_fail("do_a", "do_b"), All(DependFailure(depend_task="do_a"), DependFailure(depend_task="do_b")), id="after all fail"),

    pytest.param(after_any_finish("do_a", "do_b"), Any(DependFinish(depend_task="do_a"), DependFinish(depend_task="do_b")), id="after any finish"),
    pytest.param(after_any_success("do_a", "do_b"), Any(DependSuccess(depend_task="do_a"), DependSuccess(depend_task="do_b")), id="after any success"),
    pytest.param(after_any_fail("do_a", "do_b"), Any(DependFailure(depend_task="do_a"), DependFailure(depend_task="do_b")), id="after any fail"),
]

params_action = [
    pytest.param(failed.get_cond(), TaskFailed(task=None), id="has failed (self)"),
    pytest.param(succeeded.get_cond(), TaskSucceeded(task=None), id="has succeeded (self)"),
    pytest.param(started.get_cond(), TaskStarted(task=None), id="has started (self)"),
    pytest.param(finished.get_cond(), TaskFinished(task=None), id="has finished (self)"),

    pytest.param(failed("a_task").get_cond(), TaskFailed(task="a_task"), id="has failed"),
    pytest.param(succeeded("a_task").get_cond(), TaskSucceeded(task="a_task"), id="has succeeded"),
    pytest.param(started("a_task").get_cond(), TaskStarted(task="a_task"), id="has started"),
    pytest.param(finished("a_task").get_cond(), TaskFinished(task="a_task"), id="has finished"),

    pytest.param(started("a_task").this_minute.get_cond(), TaskStarted(task="a_task", period=TimeOfMinute()), id="has started (this minute)"),
    pytest.param(failed("a_task").this_day.get_cond(), TaskFailed(task="a_task", period=TimeOfDay()), id="has failed (this day)"),
    pytest.param(succeeded("a_task").today.between("12:00", "15:00"), TaskSucceeded(task="a_task", period=TimeOfDay("12:00", "15:00")), id="has succeeded (today between)"),
    pytest.param(finished("a_task").this_day.between("12:00", "15:00"), TaskFinished(task="a_task", period=TimeOfDay("12:00", "15:00")), id="has finished (this day between)"),
    pytest.param(started("a_task").this_week.get_cond(), TaskStarted(task="a_task", period=TimeOfWeek()), id="has started (this week)"),
    pytest.param(succeeded("a_task").this_week.before("Mon"), TaskSucceeded(task="a_task", period=TimeOfWeek(None, "Mon")), id="has succeeded (this week before)"),

    pytest.param(retry.get_cond(), Retry(n=-1), id="Retry infinite"),
    pytest.param(retry(2), Retry(n=2), id="Retry twice"),
]

params_running = [
    pytest.param(running(task="mytask"), TaskRunning(task="mytask"), id="is running"),
    pytest.param(running("10 mins"), TaskRunning(task=None, period=TimeSpanDelta(near="10 mins")), id="is running 10 mins"),
    pytest.param(running("10 mins", task="a_task"), TaskRunning(task="a_task", period=TimeSpanDelta(near="10 mins")), id="is running 10 mins (passed task)"),
]

params_schedule = [
    pytest.param(scheduler_running("10 mins"), SchedulerStarted(period=TimeSpanDelta("10 mins")), id="scheduler running (at least)"),
    pytest.param(scheduler_cycles(5), SchedulerCycles() > 5, id="scheduler cycles (at least)"),
    pytest.param(scheduler_cycles(less_than=5), SchedulerCycles() < 5, id="scheduler cycles (less than)"),
    pytest.param(scheduler_cycles(1, 5), (SchedulerCycles() > 1) < 5, id="scheduler cycles (between)"),
]

cron_like = [
    pytest.param(cron("1 2 3 4 5"), TaskRunnable(period=Cron('1', '2', '3', '4', '5')), id="cron 1 2 3 4 5"),
    pytest.param(cron(minute="1", hour="2", day_of_month="3", month="4", day_of_week="5"), TaskRunnable(period=Cron('1', '2', '3', '4', '5')), id="cron 1 2 3 4 5 (kwargs)"),

    pytest.param(crontime("1 2 3 4 5"), IsPeriod(period=Cron('1', '2', '3', '4', '5')), id="crontime 1 2 3 4 5"),
    pytest.param(crontime(minute="1", hour="2", day_of_month="3", month="4", day_of_week="5"), IsPeriod(period=Cron('1', '2', '3', '4', '5')), id="crontime 1 2 3 4 5 (kwargs)"),
]

@pytest.mark.parametrize(
    "cond,result",
    params_basic + params_time + params_task_exec + params_pipeline + params_action + params_running+ params_schedule + cron_like
)
def test_api(cond, result):
    assert cond == result

@pytest.mark.parametrize(
    "cond",
    [
        pytest.param(retry, id="Retry"),
        pytest.param(finished, id="Finished"),
        pytest.param(daily, id="Daily"),
    ]
)
def test_observe(cond, session):
    # NOTE: We test that observe does not crash (not the output)
    task = FuncTask(func=lambda: None, name="mytask", session=session)

    cond.observe(task=task, session=session)

def test_fail():
    with pytest.raises(ValueError):
        every("5 seconds", based="oops")



from datetime import datetime
from typing import ClassVar, Dict, List, Tuple, Union
from abc import abstractmethod
from dataclasses import dataclass

from rocketry.pybox.time import to_microseconds, timedelta_to_str, datetime_to_dict, to_timedelta
from .base import Any, TimeInterval

@dataclass(frozen=True, repr=False)
class AnchoredInterval(TimeInterval):
    """Base class for interval for those that have
    fixed time unit (that can be converted to microseconds).

    Converts start and end to microseconds where
    0 represents the beginning of the interval
    (ie. for a week: monday 00:00 AM) and max
    is number of microseconds from start to the
    end of the interval.

    Examples for start=0 microseconds
        - for a week: Monday
        - for a day: 00:00
        - for a month: 1st day


    Methods:
    --------
        anchor_dict --> int : Calculate corresponding microseconds from dict
        anchor_str --> int : Calculate corresponding microseconds from string
        anchor_int --> int : Calculate corresponding microseconds from integer

    Properties:
        start --> int : microseconds on the interval till the start
        end   --> int : microseconds on the interval till the end

    Class attributes:
    -----------------
        _scope [str] :
    """
    _start: int
    _end: int

    components: ClassVar[Tuple[str]] = ("year", "month", "week", "day", "hour", "minute", "second", "microsecond")

    # Components that have always fixed length (exactly the same amount of time)
    _fixed_components: ClassVar[Tuple[str]] = ("week", "day", "hour", "minute", "second", "microsecond")

    _scope: ClassVar[str] = None # Scope of the full period. Ie. day, hour, second, microsecond
    _scope_max: ClassVar[int] = None # Max in microseconds of the

    _unit_resolution: ClassVar[int] = None # Microseconds of one unit (if start/end is int)
    _unit_names: ClassVar[List] = None
    _unit_mapping: ClassVar[Dict[str, int]] = {}

    def __init__(self, start=None, end=None, time_point=None, starting=None, right_closed=False):

        if start is None and end is None:
            if time_point:
                raise ValueError("Full cycle cannot be point of time")
            object.__setattr__(self, "_start", 0)
            object.__setattr__(self, "_end", self._scope_max)
        else:
            self.set_start(start)
            self.set_end(end, time_point=time_point, right_closed=right_closed, starting=starting)

    def anchor(self, value, **kwargs):
        "Turn value to nanoseconds relative to scope of the class"
        if isinstance(value, dict):
            # {"hour": 10, "minute": 20}
            return self.anchor_dict(value, **kwargs)

        elif isinstance(value, int):
            # start is considered as unit of the second behind scope
            return self.anchor_int(value, **kwargs)

        elif isinstance(value, float):
            # start is considered as unit of the second behind scope
            return self.anchor_float(value, **kwargs)

        elif isinstance(value, str):
            return self.anchor_str(value, **kwargs)
        raise TypeError(value)

    def anchor_float(self, i, **kwargs):
        raise ValueError("Float conversion not supported")

    def anchor_int(self, i, side=None, time_point=None, **kwargs):
        if side == "end":
            return (i + 1) * self._unit_resolution
        return i * self._unit_resolution

    def anchor_dict(self, d, **kwargs):
        comps = self._fixed_components[(self._fixed_components.index(self._scope) + 1):]
        kwargs = {key: val for key, val in d.items() if key in comps}
        return to_microseconds(**kwargs)

    def anchor_dt(self, dt: datetime, **kwargs) -> int:
        "Turn datetime to nanoseconds according to the scope (by removing higher time elements)"
        components = self.components
        components = components[components.index(self._scope) + 1:]
        d = datetime_to_dict(dt)
        d = {
            key: val
            for key, val in d.items()
            if key in components
        }

        return to_microseconds(**d)

    @classmethod
    def create_range(cls, start:Union[str, int]=None, end:Union[str, int]=None, step:int=None):
        if isinstance(start, str):
            start = cls._unit_mapping[start.lower()]
        if isinstance(end, str):
            end = cls._unit_mapping[end.lower()]

        periods = tuple(
            cls.at(step)
            for step in cls._unit_names[start:end:step]
        )
        return Any(*periods)

    def set_start(self, val):
        if val is None:
            ms = 0
        else:
            ms = self.anchor(val, side="start")
        self._validate(ms, orig=val)
        object.__setattr__(self, "_start", ms)
        object.__setattr__(self, "_start_orig", val)

    def set_end(self, val, right_closed=False, time_point=False, starting=False):
        if time_point and val is None:
            # Interval is "at" type, ie. on monday, at 10:00 (10:00 - 10:59:59)
            ms = self.to_timepoint(self._start)
        elif starting and val is None:
            ms = self._start
        elif val is None:
            ms = self._scope_max
        else:
            ms = self.anchor(val, side="end", time_point=time_point)

        if right_closed:
            # We use left closed in intervals so we add one unit to make it closed
            # given the end argument, ie. if "09:00 to 10:00" excludes 10:00
            # we can include it by adding one nanosecond to 10:00
            ms += 1
        self._validate(ms, orig=val)
        object.__setattr__(self, "_end", ms)
        object.__setattr__(self, "_end_orig", val)

    def to_timepoint(self, ms:int):
        "Turn microseconds to the period's timepoint"
        # Ie. Monday --> Monday 00:00 to Monday 24:00
        # By default assumes linear scale (like week)
        # but can be overridden for non linear such as year
        end_ms = ms + self._unit_resolution
        if end_ms >= self._scope_max:
            # Over period
            end_ms = end_ms - self._scope_max
        return end_ms

    def _validate(self, n:int, orig):
        if n < 0 or n > self._scope_max:
            raise ValueError(f"Out of bound: {repr(orig)}")

    @property
    def start(self):
        delta = to_timedelta(self._start, unit="microsecond")
        repr_scope = self.components[self.components.index(self._scope) + 1]
        return timedelta_to_str(delta, default_scope=repr_scope)

    @start.setter
    def start(self, val):
        self.set_start(val)

    @property
    def end(self):
        delta = to_timedelta(self._end, unit="microsecond")
        repr_scope = self.components[self.components.index(self._scope) + 1]
        return timedelta_to_str(delta, default_scope=repr_scope)

    @end.setter
    def end(self, val):
        self.set_end(val)

    @abstractmethod
    def anchor_str(self, s, **kwargs) -> int:
        raise NotImplementedError

    def __contains__(self, dt) -> bool:
        "Whether dt is in the interval"

        ms_start = self._start
        ms_end = self._end

        if self.is_full():
            # As there is no time in between,
            # the interval is considered full
            # cycle (ie. from 10:00 to 10:00)
            return True

        ms = self.anchor_dt(dt) # In relative nanoseconds (removed more accurate than scope)

        is_over_period = ms_start > ms_end # period is overnight, over weekend etc.
        if not is_over_period:
            # Note that the period is right opened (end point excluded)
            return ms_start <= ms < ms_end
        else:
            # Note that the period is right opened (end point excluded)
            return ms >= ms_start or ms < ms_end

    def is_full(self):
        "Whether every time belongs to the period (but there is still distinct intervals)"
        return (self._start == self._end) or (self._start == 0 and self._end == self._scope_max)

    def get_scope_back(self, dt):
        "Override if offsetting back is different than forward"
        return self._scope_max

    def get_scope_forward(self, dt):
        "Override if offsetting back is different than forward"
        return self._scope_max

    def rollstart(self, dt):
        "Roll forward to next point in time that on the period"
        if dt in self:
            return dt
        else:
            return self.next_start(dt)

    def rollend(self, dt):
        "Roll back to previous point in time that is on the period"
        if dt in self:
            return dt
        else:
            return self.prev_end(dt)

    def next_start(self, dt):
        "Get next start point of the period"
        ms = self.anchor_dt(dt) # In relative nanoseconds (removed more accurate than scope)
        ms_start = self._start
        ms_end = self._end

        if ms < ms_start:
            # not in period, over night
            #                     dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end

            # not in period, earlier than start
            #          dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start

            # in period, over night, earlier than start
            #            dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end
            offset = to_timedelta(int(ms_start) - int(ms), unit="microsecond")
        else:
            # not in period, later than start
            #      dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start

            # in period, over night, later than start
            #       dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end

            # in period
            #                    dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start
            ms_scope = self.get_scope_forward(dt)
            offset = to_timedelta(int(ms_start) - int(ms) + ms_scope, unit="microsecond")
        return dt + offset

    def next_end(self, dt):
        "Get next end point of the period"
        ms = self.anchor_dt(dt) # In relative nanoseconds (removed more accurate than scope)
        ms_start = self._start
        ms_end = self._end

        if ms < ms_end:
            # in period
            #                    dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start

            # in period, over night, earlier than end
            #            dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end

            # not in period, earlier than end
            #          dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start
            offset = to_timedelta(int(ms_end) - int(ms), unit="microsecond")
        else:
            # not in period, over night
            #                     dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end

            # not in period, later than end
            #      dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start

            # in period, over night, later than end
            #       dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end
            ms_scope = self.get_scope_forward(dt)
            offset = to_timedelta(int(ms_end) - int(ms) + ms_scope, unit="microsecond")
        return dt + offset

    def prev_start(self, dt):
        "Get previous start point of the period"
        ms = self.anchor_dt(dt) # In relative nanoseconds (removed more accurate than scope)
        ms_start = self._start

        if ms < ms_start:
            # not in period, over night
            #                     dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end

            # not in period, earlier than start
            #          dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start

            # in period, over night, earlier than start
            #            dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end
            ms_scope = self.get_scope_back(dt)
            offset = to_timedelta(int(ms_start) - int(ms) - ms_scope, unit="microsecond")
        else:
            # not in period, later than start
            #      dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start

            # in period, over night, later than start
            #       dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end

            # in period
            #                    dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start
            offset = to_timedelta(int(ms_start) - int(ms), unit="microsecond")
        return dt + offset

    def prev_end(self, dt):
        "Get pervious end point of the period"
        ms = self.anchor_dt(dt) # In relative nanoseconds (removed more accurate than scope)
        ms_end = self._end

        if ms < ms_end:
            # in period
            #                    dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start

            # in period, over night, earlier than end
            #            dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end

            # not in period, earlier than end
            #          dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start
            ms_scope = self.get_scope_back(dt)
            offset = to_timedelta(int(ms_end) - int(ms) - ms_scope, unit="microsecond")
        else:
            # not in period, over night
            #                     dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end

            # not in period, later than end
            #      dt
            # --<---------->-----------<-------------->--
            #  end   |   start        end    |      start

            # in period, over night, later than end
            #       dt
            #  -->----------<----------->--------------<-
            #  start   |   end        start     |     end
            offset = to_timedelta(int(ms_end) - int(ms), unit="microsecond")

        return dt + offset

    def repr_ms(self, n:int):
        "Microseconds to representative format"
        return repr(n)

    def __repr__(self):
        cls_name = type(self).__name__
        start = self.repr_ms(self._start) if not hasattr(self, "_start_orgi") else self._start_orig
        end = self.repr_ms(self._end) if not hasattr(self, "_end_orgi") else self._end_orig
        return f'{cls_name}({start}, {end})'

    def __str__(self):
        # Hour: '15:'
        start_ms = self._start
        end_ms = self._end

        scope = self._scope
        repr_scope = self.components[self.components.index(self._scope) + 1]

        to_start = to_timedelta(start_ms, unit="microsecond")
        to_end = to_timedelta(end_ms, unit="microsecond")

        start_str = timedelta_to_str(to_start, default_scope=repr_scope)
        end_str = timedelta_to_str(to_end, default_scope=repr_scope)

        start_str = f"0 {repr_scope}s" if not start_str else start_str
        return f"{start_str} - {end_str}"

    @classmethod
    def at(cls, value):
        return cls(value, time_point=True)

    @classmethod
    def starting(cls, value):
        return cls(value, starting=True)



from functools import partial
import re

from typing import Callable, Dict, List

from rocketry.pybox.string.parse import ClosureParser
from rocketry.pybox.container.visitor import Visitor


class InstructionParser:

    def __init__(self, item_parser:Callable, operators:List[Dict[str, Callable]]):
        self.item_parser = item_parser

        self.operators = operators
        self.symbols = set([oper["symbol"] for oper in operators])

    def __call__(self, s:str, **kwargs):
        """Parse a string to condition. Allows logical operators.

        Reserved keywords:
            "&" : and operator
            "|" : or operator
            "~" : not operator
            "(" : opening closure
            ")" : closing closure

        These characters cannot be found in
        individual condition parsing (ie.
        in the names of tasks).
        """
        p = ClosureParser()
        v = Visitor(visit_types=(list,))

        # 1. Split closures
        l = p.to_list(s)

        # 2. Split operations
        v.assign_elements(l, self._split_operations)

        # 2a. Remove extra tuples
        v.apply(l, _flatten_tuples)

        v.assign_elements(l, partial(self._parse, **kwargs))

        e = v.reduce(l, self._assemble)
        return e

    def _parse(self, __s:tuple, **kwargs):
        s = __s
        def parse_string(s):
            s = s.strip()
            if s in ("&", "|", "~"):
                return s
            else:
                return self.item_parser(s, **kwargs)

        if isinstance(s, str):
            return parse_string(s)
        else:
            return tuple(parse_string(e) for e in s)

    def _assemble(self, *s:tuple):

        v = Visitor(visit_types=(list, tuple))
        s = v.flatten(s)

        #s = self._assemble_not(s)
        #s = self._assemble_and(s)
        #s = self._assemble_or(s)

        for operator in self.operators:
            oper_str = operator["symbol"]
            oper_func = operator["func"]
            oper_side = operator["side"]
            s = self._assemble_oper(s, oper_str=oper_str, oper_func=oper_func, side=oper_side)

        # TODO: Clean this mess (but be careful)

        return s[0] if isinstance(s, tuple) else s

    def _assemble_oper(self, s:list, oper_str:str, oper_func:Callable, side="both"):
        s = list(reversed(s))
        while self._contains_operator(s, oper_str):
            pos = self._index(s, [oper_str])

            # Set the comparison object to "|" in the list

            if side == "both":
                obj = oper_func(s[pos+1], s[pos-1])
                s[pos] = obj
                # NOTE: We have reversed the "s" thus we also put the arguments to

                # Remove lhs and rhs of the comparison as they are embedded in comparison object
                del s[pos-1]
                del s[pos+1-1] # We -1 because we already removed one element thus pos is misaligned
            elif side == "right":
                obj = oper_func(s[pos-1])
                s[pos-1] = obj
                del s[pos]
            elif side == "left":
                obj = oper_func(s[pos+1])
                s[pos+1] = obj
                del s[pos]

        return tuple(reversed(s))

    @staticmethod
    def _contains_operator(s:list, oper_str:str):
        for e in s:
            if isinstance(e, str) and e in (oper_str,):
                return True
        return False

    @staticmethod
    def _index(s:list, items:list):
        "Get "
        for i, e in enumerate(s):
            if isinstance(e, str) and e in items:
                return i
        raise KeyError

    def _split_operations(self, s:str):
        # The following are considered as reserved operators
        #   "&" : and operator
        #   "|" : or operator
        #   "~" : not operator
        symbols = ''.join(self.symbols)
        regex = r'([' + symbols + '])'
        s = s.strip()
        if bool(re.search(regex, s)):
            l = re.split(regex, s)
            l = [elem for elem in l if elem.strip()]
            if len(l) == 1:
                # Has only the
                return l[0]

            return tuple(l)
        else:
            return s


def _flatten_tuples(cont):

    for i, item in enumerate(cont):
        if isinstance(item, tuple):
            cont.pop(i)
            for j, tpl_item in enumerate(item):
                cont.insert(i+j, tpl_item)

    return cont


# Conditions



from rocketry.time import (
    TimeOfWeek,
    TimeOfDay,
    TimeOfHour,
    TimeOfMinute,
    TimeOfMonth
)

TIME_CLASSES = {
    "monthly": TimeOfMonth,
    "weekly": TimeOfWeek,
    "daily": TimeOfDay,
    "hourly": TimeOfHour,
    "minutely": TimeOfMinute,

    "month": TimeOfMonth,
    "week": TimeOfWeek,
    "day": TimeOfDay,
    "hour": TimeOfHour,
    "minute": TimeOfMinute,

    "this month": TimeOfMonth,
    "this week": TimeOfWeek,
    "today": TimeOfDay,
    "this hour": TimeOfHour,
    "this minute": TimeOfMinute,
}

# Utility funcs
def get_between(type_, start, end):
    type_ = type_.lower()
    cls = TIME_CLASSES[type_]
    return cls(start, end)

def get_after(type_, start):
    type_ = type_.lower()
    cls = TIME_CLASSES[type_]
    return cls(start, None)

def get_before(type_, end):
    type_ = type_.lower()
    cls = TIME_CLASSES[type_]
    return cls(None, end)

def get_on(type_, start):
    type_ = type_.lower()
    cls = TIME_CLASSES[type_]
    return cls(start, time_point=True)

def get_full_cycle(type_, start=None):
    type_ = type_.lower()
    cls = TIME_CLASSES[type_]
    return cls(start, start)



from typing import List, Optional, Union

from pydantic import BaseModel

from rocketry.conditions import Any, All, DependFinish, DependSuccess
from rocketry.conditions.task import DependFailure
from rocketry.core import Task

from rocketry import Session

class Link:

    def __init__(self,
                 parent: Task,
                 child: Task,
                 relation: Optional[Union[DependSuccess, DependFailure, DependFinish]]=None,
                 type: Optional[Union[Any, All]]=None):
        self.parent = parent
        self.child = child
        self.relation = relation
        self.type = type

    def __iter__(self):
        return iter((self.parent, self.child))

    def __eq__(self, other):
        if type(self) == type(other):
            return (
                self.parent == other.parent
                and self.child == other.child
                and self.relation == other.relation
                and self.type == other.type
            )
        return False

    def __str__(self):
        s = f'{self.parent.name!r} -> {self.child.name!r}'
        if self.type is All:
            s = s + ' (multi)'
        return s

    def __repr__(self):
        return f'Link({self.parent.name!r}, {self.child.name!r}, relation={getattr(self.relation, "__name__", None)}, type={getattr(self.type, "__name__", None)})'

class Dependencies(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    session: Session

    def __init__(self, session, **kwargs):
        super().__init__(session=session, **kwargs)

    def __iter__(self):
        for task in self.session.tasks:
            yield from self._get_links(task)

    def _get_links(self, task:Task) -> Union[Any, All]:
        cond = task.start_cond
        if isinstance(cond, (Any, All)):
            for subcond in cond:
                if isinstance(subcond, (DependFinish, DependSuccess, DependFailure)):
                    req_task = subcond.depend_task
                    req_task = self.session[req_task]
                    yield Link(parent=req_task, child=task, relation=type(subcond), type=type(cond))
        elif isinstance(cond, (DependFinish, DependSuccess, DependFailure)):
            req_task = cond.depend_task
            req_task = self.session[req_task]
            yield Link(req_task, task, relation=type(cond))


def get_dependencies(session) -> List[Link]:
    "Get list of dependency links"
    return list(Dependencies(session))



import logging
import time

import pytest

from redbird.repos import MemoryRepo
from redbird.logging import RepoHandler

from rocketry.args import TerminationFlag, Session as SessionArg
from rocketry.core import BaseCondition
from rocketry.tasks import FuncTask
from rocketry.time import TimeDelta
from rocketry.exc import TaskTerminationException
from rocketry.conditions import SchedulerCycles, SchedulerStarted, TaskStarted, AlwaysTrue
from rocketry.core import BaseArgument
from rocketry.exc import TaskSetupError

class FailingArgument(BaseArgument):

    def __init__(self, fail_in):
        self.fail_in = fail_in

    def stage(self, **kwargs):
        if self.fail_in == "stage":
            raise RuntimeError("Deliberate failure")
        return self

    def get_value(self, **kwargs):
        if self.fail_in == "get_value":
            raise RuntimeError("Deliberate failure")


class FailingCondition(BaseCondition):

    def get_state(self):
        raise RuntimeError("Deliberate failure")

def do_stuff():
    ...

def do_stuff_with_arg(arg):
    ...

def run_slow_thread(flag=TerminationFlag(), session=SessionArg()):
    while session.scheduler.n_cycles < 2 and not flag.is_set():
        time.sleep(0.0001)
    if flag.is_set():
        raise TaskTerminationException()

def run_slow():
    time.sleep(2)

@pytest.mark.parametrize(
    "execution,fail_in", [
        pytest.param("main", "get_value", id="main"),

        pytest.param("thread", "get_value", id="thread, get_value"),
        pytest.param("thread", "stage", id="thread, stage"),

        pytest.param("process", "get_value", id="process, get_value"),
        pytest.param("process", "stage", id="process, stage"),
    ]
)
def test_param_failure(tmpdir, execution, session, fail_in):
    session.config.silence_task_prerun = True # Prod setting
    task = FuncTask(do_stuff_with_arg, name="a task", parameters={"arg": FailingArgument(fail_in)}, start_cond=AlwaysTrue(), execution=execution, session=session)

    session.config.shut_cond = (TaskStarted(task="a task") >= 1) | ~SchedulerStarted(period=TimeDelta("5 second"))
    session.start()
    assert task.status == "fail"

    records = list(map(lambda d: d.dict(exclude={'created'}), task.logger.get_records()))
    assert [{"task_name": "a task", "action": "run"}, {"task_name": "a task", "action": "fail"}] == records

@pytest.mark.parametrize(
    "execution,fail_in", [
        pytest.param("main", "get_value", id="main"),

        pytest.param("thread", "get_value", id="thread, get_value"),
        pytest.param("thread", "stage", id="thread, stage"),

        pytest.param("process", "get_value", id="process, get_value"),
        pytest.param("process", "stage", id="process, stage"),
    ]
)
def test_session_param_failure(tmpdir, execution, session, fail_in):
    session.config.silence_task_prerun = True # Prod setting
    session.parameters["arg"] = FailingArgument(fail_in)

    task = FuncTask(do_stuff_with_arg, name="a task", start_cond=AlwaysTrue(), execution=execution, session=session)

    session.config.shut_cond = (TaskStarted(task="a task") >= 1) | ~SchedulerStarted(period=TimeDelta("5 second"))
    session.start()
    assert task.status == "fail"

    records = list(map(lambda d: d.dict(exclude={'created'}), task.logger.get_records()))
    assert [{"task_name": "a task", "action": "run"}, {"task_name": "a task", "action": "fail"}] == records



@pytest.mark.parametrize(
    "execution,fail_in", [
        pytest.param("main", "get_value", id="main"),

        pytest.param("thread", "get_value", id="thread, get_value", marks=pytest.mark.xfail(reason="exception relying via thread not implemented")),
        pytest.param("thread", "stage", id="thread, stage"),

        pytest.param("process", "get_value", id="process, get_value", marks=pytest.mark.xfail(reason="exception relying via queue not implemented")),
        pytest.param("process", "stage", id="process, stage"),
    ]
)
def test_raise_param_failure(execution, session, fail_in):
    session.config.silence_task_prerun = False
    task = FuncTask(do_stuff_with_arg, name="a task", parameters={"arg": FailingArgument(fail_in)}, start_cond=AlwaysTrue(), execution=execution, session=session)
    session.config.shut_cond = (TaskStarted(task="a task") >= 1) | ~SchedulerStarted(period=TimeDelta("5 second"))

    with pytest.raises(TaskSetupError):
        session.start()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_raise_task_start_cond_failure(execution, session):
    session.config.silence_cond_check = False
    task = FuncTask(do_stuff, name="a task", start_cond=FailingCondition(), execution=execution, session=session)

    session.config.shut_cond = ~SchedulerStarted(period=TimeDelta("5 second"))

    with pytest.raises(RuntimeError):
        session.start()

@pytest.mark.parametrize("execution", ["thread", "process"])
def test_raise_task_end_cond_failure(execution, session):
    session.config.silence_cond_check = False
    func = run_slow if execution == "process" else run_slow_thread if execution == "thread" else do_stuff
    task = FuncTask(func, name="a task", start_cond=AlwaysTrue(), end_cond=FailingCondition(), execution=execution, session=session)

    session.config.shut_cond = ~SchedulerStarted(period=TimeDelta("5 second"))

    with pytest.raises(RuntimeError):
        session.start()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_raise_sched_cond_failure(execution, session):
    session.config.silence_cond_check = False
    task = FuncTask(do_stuff, name="a task", execution=execution, session=session)


    session.config.shut_cond = FailingCondition()

    with pytest.raises(RuntimeError):
        session.start()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
@pytest.mark.parametrize("which", ["start_cond", "end_cond"])
def test_silence_task_cond_failure(execution, which, session):
    logger = logging.getLogger("rocketry.scheduler")
    sched_logs = MemoryRepo()
    handler = RepoHandler(sched_logs)
    try:
        logger.addHandler(handler)

        session.config.silence_cond_check = True
        session.config.shut_cond =SchedulerCycles() >= 3
        if which == "start_cond":
            task = FuncTask(do_stuff, name="a task", start_cond=FailingCondition(), execution=execution, session=session)
        elif which == "end_cond":
            func = run_slow if execution == "process" else run_slow_thread if execution == "thread" else do_stuff
            task = FuncTask(func, name="a task", start_cond=~TaskStarted(), end_cond=FailingCondition(), execution=execution, session=session)

        session.start()

        if which == "start_cond":
            assert task.status is None
            errors = sched_logs.filter_by(levelname="ERROR").all()
            assert len(errors) == 3

            assert errors[0]['msg'] == "Condition crashed for task 'a task'"
        elif which == "end_cond":
            if execution == "main":
                assert task.status == "success"
            else:
                assert task.status == "terminate"


    finally:
        logger.handlers = [
            hdlr
            for hdlr in logger.handlers
            if hdlr is not handler
        ]

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_silence_sched_cond_failure(execution, session):
    # Failing shut_cond crashes also with silence_cond_check = True
    session.config.silence_cond_check = True
    task = FuncTask(do_stuff, name="a task", execution=execution, session=session)

    session.config.shut_cond = FailingCondition()
    with pytest.raises(RuntimeError):
        session.start()



import sys

from rocketry.pybox.time.convert import to_datetime
from rocketry.testing.log import create_task_record

def log_task_record(task, now, action, start_time=None):
    "Copy of the mechanism of creating an action log"
    if action == "fail":
        try:
            raise RuntimeError("Deliberate failure")
        except:
            exc_info = sys.exc_info()
    else:
        exc_info = None

    msg = {
        "run": "",
        "fail": f"Task '{task.name}' failed",
        "success": "",
    }[action]

    now = to_datetime(now)

    if action == "run":
        start_time = now
    else:
        if start_time is None:
            start_time = task.last_run
        else:
            start_time = to_datetime(start_time)


    record = create_task_record(
        # The content here should not matter for task status
        msg=msg, args=(), exc_info=exc_info,
        created=now, action=action, taask_name=task.name,
    )

    record.start = start_time
    if action != "run":
        record.end = now
        record.runtime = now - start_time
    else:
        task._last_run = start_time

    task.logger.handle(record)
    task._status = action

    setattr(task, f"_last_{action}", now)
    



import pytest

from rocketry.conditions import (
    AlwaysTrue, AlwaysFalse,
    All, Any, Not,
    TaskFinished, TaskRunning,
)

@pytest.mark.parametrize(
    "get_cond",
    [
        pytest.param(lambda: AlwaysTrue(), id="AlwaysTrue"),
        pytest.param(lambda: AlwaysFalse(), id="AlwaysFalse"),
        pytest.param(lambda: All(AlwaysTrue(), AlwaysTrue()), id="All"),
        pytest.param(lambda: Any(AlwaysTrue(), AlwaysTrue()), id="Any"),
        pytest.param(lambda: Not(AlwaysTrue()), id="Not"),
        pytest.param(lambda: Not(Any(AlwaysTrue(), All(AlwaysTrue(), AlwaysTrue()))), id="Nested"),
        pytest.param(lambda: TaskFinished(task="mytask"), id="Comparable 1"),
        pytest.param(lambda: TaskRunning(task="mytask"), id="Comparable 2"),
    ],
)
def test_equal(get_cond):
    assert get_cond() == get_cond()


@pytest.mark.parametrize(
    "get_a,get_b",
    [
        pytest.param(
            lambda: AlwaysTrue(),
            lambda: AlwaysFalse(),
            id="AlwaysTrue/AlwaysFalse"
        ),
        pytest.param(
            lambda: All(AlwaysTrue(), AlwaysTrue()),
            lambda: Any(AlwaysTrue(), AlwaysTrue()),
            id="All/Any"
        ),
        pytest.param(
            lambda: AlwaysTrue(),
            lambda: Not(AlwaysTrue()),
            id="Not"
        ),
        pytest.param(
            lambda: TaskFinished(task="mytask"),
            lambda: TaskFinished(task="another task"),
            id="TaskFinished"
        ),
    ],
)
def test_equal_not(get_a, get_b):
    assert not get_a() == get_b()




from pathlib import Path
from textwrap import dedent
import pytest
from rocketry.args.builtin import SimpleArg
from rocketry.core.condition.base import AlwaysTrue
from rocketry.tasks import FuncTask
from rocketry.conditions import SchedulerCycles

def run_parametrized(arg):
    assert arg == "correct"

def run_parametrized_correct(arg=SimpleArg('correct')):
    assert arg == "correct"

def run_parametrized_incorrect(arg=SimpleArg("incorrect")):
    assert arg == "correct"

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_batch_favored(execution, session):
    # batch params > task params > func params > session params
    session.parameters['arg'] = 'incorrect'

    task = FuncTask(
        run_parametrized_incorrect,
        start_cond=AlwaysTrue(),
        name="task",
        execution=execution,
        session=session,
        parameters={"arg": "incorrect"}
    )
    task.run(arg="correct")
    session.config.shut_cond = SchedulerCycles() == 1
    session.start()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()

@pytest.mark.parametrize("delayed", [True, False])
@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_task_favored(execution, session, tmpdir, delayed):
    # task params > func params > session params
    session.parameters['arg'] = 'incorrect'

    if delayed:
        funcfile = tmpdir.join("script_task_favored.py")
        funcfile.write(dedent("""
            from rocketry.args import SimpleArg
            def run_parametrized_incorrect(arg=SimpleArg("incorrect")):
                assert arg == "correct"
        """))
        task = FuncTask(
            path=Path(funcfile), func_name="run_parametrized_incorrect",
            start_cond=AlwaysTrue(),
            name="task",
            execution=execution,
            session=session,
            parameters={"arg": "correct"}
        )
    else:
        task = FuncTask(
            run_parametrized_incorrect,
            start_cond=AlwaysTrue(),
            name="task",
            execution=execution,
            session=session,
            parameters={"arg": "correct"}
        )
    session.config.shut_cond = SchedulerCycles() == 1
    session.start()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_func_favored(execution, session):
    # func params > session params
    session.parameters['arg'] = 'incorrect'

    task = FuncTask(
        run_parametrized_correct,
        start_cond=AlwaysTrue(),
        name="task",
        execution=execution,
        session=session,
    )
    session.config.shut_cond = SchedulerCycles() == 1
    session.start()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_session_favored(execution, session):
    # session params
    session.parameters['arg'] = 'correct'

    task = FuncTask(
        run_parametrized,
        start_cond=AlwaysTrue(),
        name="task",
        execution=execution,
        session=session,
    )
    session.config.shut_cond = SchedulerCycles() == 1
    session.start()

    logger = task.logger
    assert 1 == logger.filter_by(action="run").count()
    assert 1 == logger.filter_by(action="success").count()



from rocketry.conditions.api import (
    true, false,

    every,
    secondly, minutely, hourly, daily, weekly, monthly,
    time_of_second, time_of_minute, time_of_hour, time_of_day, time_of_week, time_of_month,

    after_finish, after_success, after_fail,

    after_all_finish, after_all_success, after_all_fail,
    after_any_finish, after_any_success, after_any_fail,

    started, succeeded, failed, finished,
    retry,

    scheduler_running, scheduler_cycles,
    running,
    cron,
    crontime,
)



from setuptools import setup, find_packages
import versioneer

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="rocketry",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Mikael Koli",
    author_email="koli.mikael@gmail.com",
    url="https://github.com/Miksus/rocketry.git",
    packages=find_packages(),
    description="Advanced scheduling framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",

        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",

        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Office/Business :: Scheduling",

        "Topic :: System",

        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
    ],
     include_package_data=True, # for MANIFEST.in
     python_requires='>=3.7.0',

    install_requires = [
        'python-dateutil', 'redbird>=0.5.0', 'pydantic'
    ],
)



from datetime import datetime

import pytest

from rocketry.time.interval import (
    TimeOfSecond,
    TimeOfMinute,
    TimeOfHour,
    TimeOfDay,
    TimeOfWeek,
    TimeOfMonth,
)

# TimeOfSecond
# ---------

@pytest.mark.parametrize(
    "dt,time",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 11, 0, 0, 200_000),
            TimeOfSecond(200, 700),
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 11, 0, 0, 700_999),
            TimeOfSecond(200, 700),
            id="Right of interval"),
    ],
)
def test_in_time_of_second(time, dt):
    assert dt in time


@pytest.mark.parametrize(
    "dt,time",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 11, 0, 0, 199_999),
            TimeOfSecond(200, 700),
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 11, 0, 0, 800_000),
            TimeOfSecond(200, 700),
            id="Right of interval"),
    ],
)
def test_not_in_time_of_second(time, dt):
    assert dt not in time

# TimeOfMinute
# ------------

@pytest.mark.parametrize(
    "dt,time",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 11, 0, 15),
            TimeOfMinute("15.00", "45.00"),
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 11, 0, 44, 999_999),
            TimeOfMinute("15.00", "45.00"),
            id="Right of interval"),
    ],
)
def test_in_time_of_minute(time, dt):
    assert dt in time


@pytest.mark.parametrize(
    "dt,time",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 11, 0, 14, 999_999),
            TimeOfMinute("15.00", "45.00"),
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 11, 00, 45, 0),
            TimeOfMinute("15.00", "45.00"),
            id="Right of interval"),
    ],
)
def test_not_in_time_of_minute(time, dt):
    assert dt not in time

# TimeOfHour
# ----------

@pytest.mark.parametrize(
    "dt,time",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 11, 15),
            TimeOfHour("15:00", "45:00"),
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 11, 44, 59, 999_999),
            TimeOfHour("15:00", "45:00"),
            id="Right of interval"),
    ],
)
def test_in_time_of_hour(time, dt):
    assert dt in time

@pytest.mark.parametrize(
    "dt,time",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 11, 14, 59, 999_999),
            TimeOfHour("15:00", "45:00"),
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 11, 45),
            TimeOfHour("15:00", "45:00"),
            id="Right of interval"),
    ],
)
def test_not_in_time_of_hour(time, dt):
    assert dt not in time

# TimeOfDay
# ---------

@pytest.mark.parametrize(
    "dt,start,end",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 11, 00),
            "10:00", "12:00",
            id="Middle of interval"),

        # Left is closed
        pytest.param(
            datetime(2020, 1, 1, 10, 00),
            "10:00", "12:00",
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 1, 22, 00),
            "22:00", "02:00",
            id="Left of overnight interval"),

        # Overnight
        pytest.param(
            datetime(2020, 1, 1, 23, 59, 59, 999999),
            "22:00", "02:00",
            id="Middle left of overnight interval"),
        pytest.param(
            datetime(2020, 1, 1, 00, 00),
            "22:00", "02:00",
            id="Middle right of overnight interval"),

        # Full Cycle
        pytest.param(
            datetime(2020, 1, 1, 10, 00),
            None, None,
            id="Full interval"),
        pytest.param(
            datetime(2020, 1, 1, 10, 00),
            "10:00", "10:00",
            id="Joint of full interval"),
        pytest.param(
            datetime(2020, 1, 1, 12, 00),
            "10:00", "10:00",
            id="Right of full interval"),
        pytest.param(
            datetime(2020, 1, 1, 8, 00),
            "10:00", "10:00",
            id="Left of full interval"),
    ],
)
def test_in_time_of_day(start, end, dt):
    time = TimeOfDay(start, end)
    assert dt in time


@pytest.mark.parametrize(
    "dt,start,end",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 9, 59, 59, 999999),
            "10:00", "12:00",
            id="Left from interval"),
        pytest.param(
            datetime(2020, 1, 1, 12, 00, 00, 1),
            "10:00", "12:00",
            id="Right from interval"),

        # Right is opened
        pytest.param(
            datetime(2020, 1, 1, 12, 00),
            "10:00", "12:00",
            id="Right of interval"),
        pytest.param(
            datetime(2020, 1, 1, 2, 00),
            "22:00", "02:00",
            id="Right of overnight interval"),

        # Overnight
        pytest.param(
            datetime(2020, 1, 1, 21, 59, 59, 999999),
            "22:00", "02:00",
            id="Left from overnight interval"),
        pytest.param(
            datetime(2020, 1, 1, 2, 00, 00, 1),
            "22:00", "02:00",
            id="Right from overnight interval"),
    ],
)
def test_not_in_time_of_day(start, end, dt):
    time = TimeOfDay(start, end)
    assert dt not in time


# TimeOfWeek
# ----------

# TimeOfWeek
# Year 2024 was chosen as it starts on monday
@pytest.mark.parametrize(
    "dt,start,end",
    [
        # Regular
        pytest.param(
            datetime(2024, 1, 2, 10, 00),
            "Tue 10:00", "Sat 12:00",
            id="Left of interval (with time)"),
        pytest.param(
            datetime(2024, 1, 4, 11, 00),
            "Tue 10:00", "Sat 12:00",
            id="Middle of interval (with time)"),
        pytest.param(
            datetime(2024, 1, 6, 11, 59, 59, 999999),
            "Tue 10:00", "Sat 12:00",
            id="Right of interval (with time)"),
        pytest.param(
            datetime(2024, 1, 6, 23, 59, 59, 999999),
            "Tue", "Sat",
            id="Right of interval"),

        # Over weekend
        pytest.param(
            datetime(2024, 1, 6, 10, 00),
            "Sat 10:00", "Tue 12:00",
            id="Left of over weekend interval (with time)"),
        pytest.param(
            datetime(2024, 1, 7, 23, 59, 59, 999999),
            "Sat 10:00", "Tue 12:00",
            id="Middle left of over weekend interval (with time)"),
        pytest.param(
            datetime(2024, 1, 8, 00, 00),
            "Sat 10:00", "Tue 12:00",
            id="Middle right of over weekend interval (with time)"),

        # Full Cycle
        pytest.param(
            datetime(2024, 1, 1, 00, 00),
            None, None,
            id="Full interval"),
        pytest.param(
            datetime(2024, 1, 2, 00, 00),
            "Tue", "Tue",
            id="Joint of full interval"),
        pytest.param(
            datetime(2024, 1, 1, 00, 00),
            "Tue 00:00", "Tue 00:00",
            id="Right of full interval"),
        pytest.param(
            datetime(2024, 1, 6, 00, 00),
            "Tue 00:00", "Tue 00:00",
            id="Left of full interval"),
    ],
)
def test_in_time_of_week(start, end, dt):
    time = TimeOfWeek(start, end)
    assert dt in time


@pytest.mark.parametrize(
    "dt,start,end",
    [
        # Regular
        pytest.param(
            datetime(2024, 1, 1, 0, 0),
            "Tue 10:00", "Sat 12:00",
            id="Left from interval"),
        pytest.param(
            datetime(2024, 1, 7, 14, 00, 00, 1),
            "Tue 10:00", "Sat 12:00",
            id="Right from interval"),

        # Right is opened
        pytest.param(
            datetime(2024, 1, 6, 12, 00),
            "Tue 10:00", "Sat 12:00",
            id="Right of interval"),
        pytest.param(
            datetime(2024, 1, 9, 12, 00),
            "Sat 10:00", "Tue 12:00",
            id="Right of over weekend interval"),

        # Over weekend
        pytest.param(
            datetime(2024, 1, 6, 9, 59, 59, 999999),
            "Sat 10:00", "Tue 12:00",
            id="Left from over weekend interval"),
        pytest.param(
            datetime(2024, 1, 9, 12, 1, 00),
            "Sat 10:00", "Tue 12:00",
            id="Right from over weekend interval"),
    ],
)
def test_not_in_time_of_week(start, end, dt):
    time = TimeOfWeek(start, end)
    assert dt not in time


# TimeOfMonth
# ---------

@pytest.mark.parametrize(
    "dt,time",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 2, 11, 00),
            TimeOfMonth("2.", "3."),
            id="Middle of interval"),

        # Left is closed
        pytest.param(
            datetime(2020, 1, 2, 0, 00),
            TimeOfMonth("2.", "3."),
            id="Left of interval"),
        pytest.param(
            datetime(2020, 1, 3, 23, 59, 59, 999999),
            TimeOfMonth("2.", "3."),
            id="Left of overnight interval"),

        # Over month
        pytest.param(
            datetime(2020, 1, 23),
            TimeOfMonth("22.", "2."),
            id="Middle left of over month"),
        pytest.param(
            datetime(2020, 1, 1, 00, 00),
            TimeOfMonth("22.", "2."),
            id="Middle right of over month"),

        # Full Cycle
        pytest.param(
            datetime(2020, 1, 31),
            TimeOfMonth(None, None),
            id="Full interval"),
        pytest.param(
            datetime(2020, 1, 5),
            TimeOfMonth.starting("5."),
            id="Joint of full interval"),
        pytest.param(
            datetime(2020, 1, 1),
            TimeOfMonth.starting("5."),
            id="Right of full interval"),
        pytest.param(
            datetime(2020, 1, 6),
            TimeOfMonth.starting("5."),
            id="Left of full interval"),
    ],
)
def test_in_time_of_month(time, dt):
    assert dt in time


@pytest.mark.parametrize(
    "dt,time",
    [
        # Regular
        pytest.param(
            datetime(2020, 1, 1, 23, 59, 59, 999999),
            TimeOfMonth("2.", "3."),
            id="Left from interval"),
        pytest.param(
            datetime(2020, 1, 4),
            TimeOfMonth("2.", "3."),
            id="Right from interval"),

        # Right is opened
        pytest.param(
            datetime(2020, 1, 21),
            TimeOfMonth("22.", "2."),
            id="Right of over month"),
        pytest.param(
            datetime(2020, 1, 3),
            TimeOfMonth("22.", "2."),
            id="Left of over month"),
    ],
)
def test_not_in_time_of_month(time, dt):
    assert dt not in time



from typing import Dict, List, Callable, Tuple
import inspect

class _Hooker:
    # No, this is not what you think.

    def __init__(self, hooks:List[Callable], args:Tuple=None, kwargs:Dict=None):
        self.args = () if args is None else args
        self.kwargs = {} if kwargs is None else kwargs
        self.hooks = hooks

    def prerun(self, *args, **kwargs):
        self._post_hooks = []
        for hook in self.hooks:
            result = hook(*args, **kwargs)
            if inspect.isgeneratorfunction(hook):
                gener = result
                next(gener, None) # Executes first yield
                self._post_hooks.append(gener)

    def postrun(self, *args):
        for gener in self._post_hooks:
            try:
                gener.send(args)
            except StopIteration:
                pass

def clear_hooks():
    "Remove all hooks."
    from rocketry.core import Task, Scheduler
    Task.init_hooks = []
    Scheduler.startup_hooks = []
    Scheduler.cycle_hooks = []
    Scheduler.shutdown_hooks = []



from datetime import timedelta, datetime
import pytest
from rocketry.pybox.time import to_timedelta, to_datetime, timedelta_to_str

@pytest.mark.parametrize("s,expected",
    [
        ('00:00:00', timedelta()),
        ('06:05:01', timedelta(hours=6, minutes=5, seconds=1)),
        ('06:05:01.00003', timedelta(hours=6, minutes=5, seconds=1, microseconds=30)),
        ('06:05:01.5', timedelta(hours=6, minutes=5, seconds=1, milliseconds=500)),

        ('1 days 16:05:01.00003', timedelta(days=1, hours=16, minutes=5, seconds=1, microseconds=30)),

        ('10m 20s', timedelta(minutes=10, seconds=20)),
        ('1d 5h 10m 20s', timedelta(days=1, hours=5, minutes=10, seconds=20)),
        ('1d, 5h, 10m, 20s', timedelta(days=1, hours=5, minutes=10, seconds=20)),
        ('5 hr 10 min 20 sec', timedelta(hours=5, minutes=10, seconds=20)),
        ('5 hrs 10 mins 20 secs', timedelta(hours=5, minutes=10, seconds=20)),
        ('1d 5h 10m 20s', timedelta(days=1, hours=5, minutes=10, seconds=20)),
        ("1days 5hours 10minutes 20seconds", timedelta(days=1, hours=5, minutes=10, seconds=20)),
        ("1 day 5 hour 10 minute 20 second", timedelta(days=1, hours=5, minutes=10, seconds=20)),
        ("1 days 5 hours 10 minutes 20 seconds", timedelta(days=1, hours=5, minutes=10, seconds=20)),
        ("-1 days 5 hours 10 minutes 20 seconds", -timedelta(days=1, hours=5, minutes=10, seconds=20)),
        ("--1 days 5 hours 10 minutes 20 seconds", timedelta(days=1, hours=5, minutes=10, seconds=20)),
        ("+1 days 5 hours 10 minutes 20 seconds", timedelta(days=1, hours=5, minutes=10, seconds=20)),
    ]
)
def test_timedelta_from_string(s, expected):
    assert to_timedelta(s) == expected

@pytest.mark.parametrize("unit,n,expected",
    [
        pytest.param('ns', 0, timedelta(), id="0 ns"),
        pytest.param('ns', 1e+9, timedelta(seconds=1), id="nanoseconds"),
        pytest.param('μs', 1e+6, timedelta(seconds=1), id="microseconds"),
        pytest.param('ms', 1000, timedelta(seconds=1), id="milliseconds"),
        pytest.param('s', 1, timedelta(seconds=1), id="seconds"),
        pytest.param('m', 1, timedelta(minutes=1), id="minutes"),
        pytest.param('h', 1, timedelta(hours=1), id="hours"),
    ]
)
def test_timedelta_from_int(unit, n, expected):
    assert to_timedelta(n, unit=unit) == expected

def test_timedelta_from_timedelta():
    assert to_timedelta(timedelta(hours=5)) == timedelta(hours=5)

@pytest.mark.parametrize("obj",
    [
        pytest.param(datetime(2022, 1, 1,), id="datetime"),
        pytest.param(datetime, id="class"),
    ]
)
def test_timedelta_fail(obj):
    with pytest.raises(TypeError):
        to_timedelta(obj)

@pytest.mark.parametrize("dt,kwargs,expected",
    [
        pytest.param(
            timedelta(
                days=400, hours=12, minutes=30, seconds=59, milliseconds=500, microseconds=500
            ), {}, "400 days, 12 hours, 30 minutes, 59 seconds, 500 milliseconds, 500 microseconds",
            id="default"
        ),
        pytest.param(
            timedelta(
                days=435, hours=12, minutes=30, seconds=59, milliseconds=500, microseconds=500
            ), {"include": "all"}, "1 years, 2 months, 1 weeks, 3 days, 12 hours, 30 minutes, 59 seconds, 500 milliseconds, 500 microseconds",
            id="default (all)"
        ),
        pytest.param(
            timedelta(
                days=2,
            ), {}, "2 days",
            id="default (only days)"
        ),
        pytest.param(
            timedelta(
                days=2, seconds=30,
            ), {}, "2 days, 0 hours, 0 minutes, 30 seconds",
            id="default (only days and seconds)"
        ),
        pytest.param(
            timedelta(
                days=2, hours=12, minutes=30, seconds=59, milliseconds=500, microseconds=500
            ), {"format": "short"}, "2d, 12h, 30m, 59s, 500ms, 500μs",
            id="short"
        ),
        pytest.param(
            timedelta(
                days=2, hours=12, minutes=30, seconds=59, milliseconds=500, microseconds=500
            ), {"format": "semishort"}, "2 days, 12 hrs, 30 mins, 59 secs, 500 ms, 500 μs",
            id="semishort"
        ),
        pytest.param(
            timedelta(
                days=16, hours=6
            ),
            {"include": ("weeks", "days", "hours")},
            "2 weeks, 2 days, 6 hours",
            id="include week"
        ),
        pytest.param(
            timedelta(
                days=62, hours=6
            ),
            {"include": ("months", "days", "hours"), "days_in_month": 30},
            "2 months, 2 days, 6 hours",
            id="include month"
        ),
        pytest.param(
            timedelta(
                days=365*2 + 2, hours=6
            ),
            {"include": ("years", "days", "hours"), "days_in_year": 365},
            "2 years, 2 days, 6 hours",
            id="include year"
        ),
    ]
)
def test_timedelta_to_str(dt, kwargs, expected):
    assert timedelta_to_str(dt, **kwargs) == expected

def test_to_datetime_error():
    with pytest.raises(TypeError):
        to_datetime(sum)



import logging
from rocketry.pybox.time import to_datetime
from rocketry.core import Task

def create_record(
    level,
    pathname="",
    lineno=1,
    name="",
    msg="",
    args=None,
    created=None,
    exc_info=None,
    **kwargs
):
    if created is not None:
        if isinstance(created, str):
            created = to_datetime(created)
        if isinstance(created, float):
            created = int(created)
        if not isinstance(created, int):
            created = int(created.timestamp())

    r = logging.LogRecord(
        level=level,
        exc_info=exc_info,
        # These should not matter:
        name=name,
        pathname=pathname,
        lineno=lineno,
        msg=msg,
        args=args,
        **kwargs
    )
    # Copy-pasted from logging.LogRecord.__init__
    if created is not None:
        r.created = created
        r.msecs = (r.created - int(r.created)) * 1000
        r.relativeCreated = (r.created - logging._startTime) * 1000
    return r

def create_task_record(*args, task_name:str, action:str, **kwargs):
    """Create a task record (for testing)"""
    all_actions = Task._actions
    if action not in all_actions:
        raise ValueError(f"Allowed actions: {all_actions}")
    if 'name' not in kwargs:
        kwargs['name'] = "rocketry.task"

    r = create_record(*args, level=logging.INFO if action != "fail" else logging.ERROR, **kwargs)
    r.task_name = task_name
    r.action = action
    return r



from typing import Optional

from redbird.oper import in_, greater_equal

from rocketry.core.condition import BaseCondition
from rocketry.time.construct import get_before, get_between, get_full_cycle, get_after, get_on
from rocketry.args import Task, Session
from rocketry.core.time.utils import get_period_span
from rocketry.core.time import TimeDelta
from .utils import DependMixin, TaskStatusMixin

from ..time import IsPeriod


class TaskStarted(TaskStatusMixin):

    """Condition for whether a task has started
    (for given period).

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("task 'mytask' has started today")
    TaskStarted(task='mytask', period=TimeOfDay(None, None))
    """
    _action = 'run'

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        period = self.period
        task = self.task
        task_name = getattr(task, 'name', str(task))
        period = '' if period is None else f' {period}'
        return f"task '{task_name}' started{period}"


class TaskFailed(TaskStatusMixin):
    """Condition for whether the given task has failed
    (in given period).


    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("task 'mytask' has failed today between 10:00 and 14:00")
    TaskFailed(task='mytask', period=TimeOfDay('10:00', '14:00'))
    """
    _action = 'fail'

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        period = self.period
        task = self.task
        task_name = getattr(task, 'name', str(task))
        period = '' if period is None else f' {period}'
        return f"task '{task_name}' failed{period}"


class TaskTerminated(TaskStatusMixin):
    """Condition for whether the given task has terminated
    (in given period).


    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("task 'mytask' has terminated this week after Monday")
    TaskTerminated(task='mytask', period=TimeOfWeek('Monday', None))
    """
    _action = 'terminate'
    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        period = self.period
        task = self.task
        task_name = getattr(task, 'name', str(task))
        period = '' if period is None else f' {period}'
        return f"task '{task_name}' terminated{period}"


class TaskSucceeded(TaskStatusMixin):
    """Condition for whether the given task has succeeded
    (in given period).


    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("task 'mytask' has succeeded this month")
    TaskSucceeded(task='mytask', period=TimeOfMonth(None, None))
    """
    _action = 'success'

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        period = self.period
        task = self.task
        task_name = getattr(task, 'name', str(task))
        period = '' if period is None else f' {period}'
        return f"task '{task_name}' succeeded{period}"


class TaskFinished(TaskStatusMixin):
    """Condition for whether the given task has finished
    (in given period).


    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("task 'mytask' has finished today")
    TaskFinished(task='mytask', period=TimeOfDay(None, None))
    """
    _action = ["success", "fail", "terminate"]

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        period = self.period
        task = self.task
        task_name = getattr(task, 'name', str(task))
        period = '' if period is None else f' {period}'
        return f"task '{task_name}' finished" + period


class TaskRunning(BaseCondition):

    """Condition for whether a task is currently
    running.

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("task 'mytask' is running")
    TaskRunning(task='mytask')
    """

    def __init__(self, task=None, period:TimeDelta=None):
        self.task = task
        self.period = period
        super().__init__()

    def get_state(self, task=Task(), session=Session()):
        task = session[self.task] if self.task is not None else task
        is_running = task.is_running
        if not is_running:
            # Not running so always false
            return False
        elif is_running and self.period is None:
            # Is running (and no limit on when it stated)
            return True
        else:
            # Is running but not yet sure if the period is fulfilled
            last_run = task.get_last_run()
            start, end = get_period_span(self.period)
            return start <= last_run <= end

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        task = self.task
        task_name = getattr(task, 'name', str(task))
        return f"task '{task_name}' is running"


class TaskInacted(TaskStatusMixin):
    """Condition for whether the given task has inacted
    (in given period).

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("task 'mytask' has inacted")
    TaskInacted(task='mytask')
    """
    _action = 'inaction'

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        task = self.task
        task_name = getattr(task, 'name', str(task))
        return f"task '{task_name}' inacted"


class TaskExecutable(BaseCondition):
    """Condition for checking whether a given
    task has not finished (for given period).
    Useful to set the given task to run once
    in given period.

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("daily between 10:00 and 15:00")
    TaskExecutable(task=None, period=TimeOfDay('10:00', '15:00'))

    """

    def __init__(self, retries=None, task=None, period=None):
        self.retries = retries
        self.period = period
        self.task = task
        super().__init__()

        # TODO: If constant launching (allow launching alive tasks)
        # is to be implemented, there should be one more condition:
        # self._is_not_running

        # TODO: How to consider termination? Probably should be considered as failures without retries
        # NOTE: inaction is not considered at all

    def get_state(self, task=Task(), session=Session()):
        task = self.task if self.task is not None else task
        period = self.period
        retries = 0 if self.retries is None else self.retries

        # Form the sub statements
        has_not_succeeded = TaskSucceeded(period=period, task=task) == 0
        has_not_inacted = TaskInacted(period=period, task=task) == 0
        has_not_failed = TaskFailed(period=period, task=task) <= retries
        has_not_terminated = TaskTerminated(period=period, task=task) == 0

        isin_period = (
            # TimeDelta has no __contains__. One cannot say whether now is "past 2 hours".
            #   And please tell why this does not raise an exception then? - Future me
            #   Because the period is used in the sub statements and TimeDelta is still accepted - Senior me
            True
            if isinstance(period, TimeDelta)
            else IsPeriod(period=period).observe()
        )

        return (
            isin_period
            and has_not_inacted.observe(task=task, session=session)
            and has_not_succeeded.observe(task=task, session=session)
            and has_not_failed.observe(task=task, session=session)
            and has_not_terminated.observe(task=task, session=session)
        )

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        task = self.task
        period = self.period
        return f"task '{task}' {self.period}"

    @classmethod
    def _from_period(cls, span_type=None, **kwargs):
        period_func = {
            "between": get_between,
            "after": get_after,
            "before": get_before,
            "starting": get_full_cycle,
            None: get_full_cycle,
            "every": TimeDelta,
            "on": get_on,
        }[span_type]
        period = period_func(**kwargs)
        return cls(period=period)


class TaskRunnable(BaseCondition):
    """Condition for checking whether a given
    task has not run (for given period).
    Useful to set the given task to run once
    in given period.
    """

    def __init__(self, task=None, period=None):
        self.period = period
        self.task = task
        super().__init__()

    def get_state(self, task=Task(), session=Session()):
        task = self.task if self.task is not None else task
        period = self.period

        has_not_run = TaskStarted(period=period, task=task) == 0

        isin_period = (
            True
            if isinstance(period, TimeDelta)
            else IsPeriod(period=period).observe()
        )

        return (
            isin_period
            and has_not_run.observe(task=task, session=session)
        )


class DependFinish(DependMixin):
    """Condition for checking whether a given
    task has not finished after running a dependent
    task. Useful to set the given task to run after
    another task has finished.

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("after task 'other' finished")
    DependFinish(task=None, depend_task='other')
    """

    _dep_actions = ['success', 'fail']

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        task = self.task
        depend_task = self.depend_task
        task_name = getattr(task, 'name', str(task))
        depend_task_name = getattr(depend_task, 'name', str(depend_task))
        return f"task '{depend_task_name}' finished before '{task_name}' started"


class DependSuccess(DependMixin):
    """Condition for checking whether a given
    task has not succeeded after running a dependent
    task. Useful to set the given task to run after
    another task has succeeded.

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("after task 'other' succeeded")
    DependSuccess(task=None, depend_task='other')

    """

    _dep_actions = ['success']

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        task = self.task
        depend_task = self.depend_task
        task_name = getattr(task, 'name', str(task))
        depend_task_name = getattr(depend_task, 'name', str(depend_task))
        return f"task '{depend_task_name}' succeeded before '{task_name}' started"


class DependFailure(DependMixin):
    """Condition for checking whether a given
    task has not failed after running a dependent
    task. Useful to set the given task to run after
    another task has failed.

    Examples
    --------

    **Parsing example:**

    >>> from rocketry.parse import parse_condition
    >>> parse_condition("after task 'other' failed")
    DependFailure(task=None, depend_task='other')
    """

    _dep_actions = ['fail']

    def __str__(self):
        if hasattr(self, "_str"):
            return self._str
        task = self.task
        depend_task = self.depend_task
        task_name = getattr(task, 'name', str(task))
        depend_task_name = getattr(depend_task, 'name', str(depend_task))
        return f"task '{depend_task_name}' failed before '{task_name}' started"

class Retry(BaseCondition):
    """Condition for retrying failed attempts"""
    def __init__(self, n: Optional[int] = 1):
        self.n = int(n) if n is not None else -1
        super().__init__()

    def get_state(self, task=Task(), session=Session()):
        if self.n == 0:
            return False
        elif task.get_status() != "fail":
            # Previously did not fail, no need to retry
            return False
        elif self.n == -1:
            # Infinite retries
            return True

        last_non_fail = task.logger.filter_by(
            action=in_(['success', 'crash', 'inaction', 'terminate'])
        ).last()

        last_non_fail_created = 0 if last_non_fail is None else last_non_fail.created

        n_failed_in_row = task.logger.filter_by(
            created=greater_equal(last_non_fail_created),
            action='fail'
        ).count()
        return self.n >= n_failed_in_row



import datetime

import pytest

from rocketry.core.time.base import (
    All, Any
)
from rocketry.time.interval import TimeOfDay

from_iso = datetime.datetime.fromisoformat

@pytest.mark.parametrize(
    "dt,periods",
    [
        # Regular
        pytest.param(
            from_iso("2020-01-01 13:00:00"),
            [
                # Valid range should be 12:00 - 14:00
                TimeOfDay("08:00", "18:00"), # Dominant
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("12:00", "16:00"),
            ],
            id="Combination (center)"),
    ],
)
def test_all_in(dt, periods):
    time = All(*periods)
    assert dt in time

@pytest.mark.parametrize(
    "dt,periods",
    [
        # Regular
        pytest.param(
            from_iso("2020-01-01 11:00:00"),
            [
                # Valid range should be 12:00 - 14:00
                TimeOfDay("08:00", "18:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("12:00", "16:00"),
            ],
            id="Combination (partial outside, left)"),
        pytest.param(
            from_iso("2020-01-01 15:00:00"),
            [
                # Valid range should be 12:00 - 14:00
                TimeOfDay("08:00", "18:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("12:00", "16:00"),
            ],
            id="Combination (partial outside, right)"),
    ],
)
def test_all_not_in(dt, periods):
    time = All(*periods)
    assert dt not in time


@pytest.mark.parametrize(
    "dt,periods",
    [
        pytest.param(
            from_iso("2020-01-01 08:00:00"),
            [
                TimeOfDay("08:00", "10:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("14:00", "16:00"),
            ],
            id="Combination (left edge)"),
        pytest.param(
            from_iso("2020-01-01 11:00:00"),
            [
                TimeOfDay("08:00", "10:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("14:00", "16:00"),
            ],
            id="Combination (center)"),
        pytest.param(
            from_iso("2020-01-01 15:59:59"),
            [
                TimeOfDay("08:00", "10:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("14:00", "16:00"),
            ],
            id="Combination (right edge)"),
    ],
)
def test_any_in(dt, periods):
    time = Any(*periods)
    assert dt in time

@pytest.mark.parametrize(
    "dt,periods",
    [
        # Regular
        pytest.param(
            from_iso("2020-01-01 07:00:00"),
            [
                TimeOfDay("08:00", "10:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("14:00", "16:00"),
            ],
            id="Combination (left)"),
        pytest.param(
            from_iso("2020-01-01 17:00:00"),
            [
                TimeOfDay("08:00", "10:00"),
                TimeOfDay("10:00", "14:00"),
                TimeOfDay("14:00", "16:00"),
            ],
            id="Combination (right)"),
    ],
)
def test_any_not_in(dt, periods):
    time = Any(*periods)
    assert dt not in time



import datetime
from typing import Optional
from pydantic import BaseModel, Field

class MinimalRecord(BaseModel):
    """A log record with minimal number of fields for Rocketry to work"""
    task_name: str
    action: str = Field(description="Scheduler action: 'run', 'success', 'fail'")
    created: float

class LogRecord(MinimalRecord):
    """A logging record

    See attributes: https://docs.python.org/3/library/logging.html#logrecord-attributes
    """
    name: str
    msg: str
    levelname: str
    levelno: int
    pathname: str
    filename: str
    module: str
    exc_text: Optional[str] = Field(description="Exception text")
    lineno: int
    funcName: str
    created: float
    msecs: float
    relativeCreated: float
    thread: int
    threadName: str
    processName: str
    process: int
    message: str

    formatted_message: str = Field(description="Formatted message. This field is created by RepoHandler.")

class TaskLogRecord(MinimalRecord):

    start: Optional[datetime.datetime]
    end: Optional[datetime.datetime]
    runtime: Optional[datetime.timedelta]

    message: str
    exc_text: Optional[str]



import logging

from rocketry import Rocketry
from rocketry.conditions.task.task import TaskStarted

def set_logging_defaults():
    task_logger = logging.getLogger("rocketry.task")
    task_logger.handlers = []
    task_logger.setLevel(logging.WARNING)

def test_init_args_in_cond(session, tmpdir):
    set_logging_defaults()

    app = Rocketry(config={'task_execution': 'main'})

    @app.cond()
    def file_exists(file):
        if file == "exists.txt":
            return True
        if file == "non_existent.txt":
            return False
        raise

    # Creating some tasks
    @app.task(file_exists("exists.txt"))
    def do_always():
        ...

    @app.task(file_exists("non_existent.txt"))
    def do_never():
        ...

    app.session.config.shut_cond = TaskStarted(task=do_always)
    app.run()
    # Assert and test tasks
    assert app.session[do_always].status == "success"
    assert app.session[do_never].status is None

    assert app.session[do_always].logger.filter_by().count() >= 2
    assert app.session[do_never].logger.filter_by().count() == 0



import pytest

from rocketry.args import Return
from rocketry.conditions.scheduler import SchedulerCycles
from rocketry.core import Parameters
from rocketry.tasks import FuncTask
from rocketry.conditions import TaskStarted


def func_with_arg(value):
    pass

def func_x_with_return():
    return "x"

def func_x_with_arg(myparam):
    assert myparam == "x"


@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_normal(session, execution):

    task_return = FuncTask(
        func_x_with_return,
        name="return task",
        start_cond="~has started",
        execution=execution,
        session=session
    )
    task_return.run()
    task = FuncTask(
        func_x_with_arg,
        name="a task",
        start_cond="after task 'return task'",
        parameters={"myparam": Return('return task')},
        execution=execution,
        session=session
    )

    assert task.status is None

    session.config.shut_cond = TaskStarted(task="a task") >= 1
    session.start()

    assert dict(session.returns) == {task_return: "x", task: None}
    assert "success" == task_return.status
    assert "success" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_normal_pass_task(session, execution):

    task_return = FuncTask(
        func_x_with_return,
        name="return task",
        start_cond="~has started",
        execution=execution,
        session=session
    )
    task_return.run()
    task = FuncTask(
        func_x_with_arg,
        name="a task",
        start_cond="after task 'return task'",
        parameters={"myparam": Return(task_return)},
        execution=execution,
        session=session
    )

    assert task.status is None

    session.config.shut_cond = TaskStarted(task="a task") >= 1
    session.start()

    assert dict(session.returns) == {task_return: "x", task: None}
    assert "success" == task_return.status
    assert "success" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_normal_pass_func(session, execution):
    # This use case is for using the app pattern
    @FuncTask(
        name="return task",
        start_cond="~has started",
        execution="main",
        batches=[Parameters()],
        session=session
    )
    def func_with_arg_local():
        return 'x'

    task = FuncTask(
        func_x_with_arg,
        name="a task",
        start_cond="after task 'return task'",
        parameters={"myparam": Return(func_with_arg_local)},
        execution=execution,
        session=session,
    )

    assert task.status is None

    session.config.shut_cond = (TaskStarted(task="a task") >= 1) | (SchedulerCycles() >= 5)
    session.start()

    task_return = session["return task"]
    assert dict(session.returns) == {task_return: "x", task: None}
    assert "success" == task_return.status
    assert "success" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_missing(session, execution):
    session.config.silence_task_prerun = True # Default in prod

    task = FuncTask(
        func_with_arg,
        parameters={"myparam": Return('return task')},
        name="a task",
        execution=execution,
        session=session
    )
    task.run()
    assert task.status is None
    session.config.shut_cond = TaskStarted(task="a task") >= 1
    session.start()

    assert "fail" == task.status

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_default(session, execution):
    task = FuncTask(
        func_with_arg,
        name="return task",
        session=session
    )
    task = FuncTask(
        func_x_with_arg,
        parameters={"myparam": Return('return task', default="x")},
        name="a task",
        execution=execution,
        session=session
    )
    task.run()

    assert task.status is None
    session.config.shut_cond = TaskStarted(task="a task") >= 1
    session.start()

    assert "success" == task.status



import logging
from typing import List, Optional, Pattern, Union
import warnings

from redbird import BaseRepo
from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo
from rocketry.log.log_record import LogRecord

from rocketry.conditions import FuncCond
from rocketry.parameters import FuncParam
from rocketry import Session

class _AppMixin:

    session: Session

    def task(self, start_cond=None, name=None, **kwargs):
        "Create a task"
        return self.session.create_task(start_cond=start_cond, name=name, **kwargs)

    def param(self, name:Optional[str]=None):
        "Set one session parameter (decorator)"
        return FuncParam(name, session=self.session)

    def cond(self, syntax: Union[str, Pattern, List[Union[str, Pattern]]]=None):
        "Create a condition (decorator)"
        return FuncCond(syntax=syntax, session=self.session, decor_return_func=False)

    def params(self, **kwargs):
        "Set session parameters"
        self.session.parameters.update(kwargs)

    def include_grouper(self, group:'Grouper'):
        for task in group.session.tasks:
            if group.prefix:
                task.name = group.prefix + task.name
            if group.start_cond is not None:
                task.start_cond = task.start_cond & group.start_cond
            task.execution = group.execution if task.execution is None else task.execution

            self.session.add_task(task)
        self.session.parameters.update(group.session.parameters)

class Rocketry(_AppMixin):
    """Rocketry scheduling application"""

    def __init__(self, session:Session=None, logger_repo:Optional[BaseRepo]=None, execution=None, **kwargs):

        config = {}
        config.update(kwargs.pop('config', {}))
        if execution is not None:
            config['task_execution'] = execution

        self.session = session if session is not None else Session(**kwargs, config=config)

        logger = logging.getLogger(self.session.config.task_logger_basename)
        logger.setLevel(logging.INFO)

        self._set_logger_with_repo(logger_repo)

    def run(self, debug=False):
        "Run the scheduler"
        self.session.config.debug = debug
        self.session.set_as_default()
        self.session.start()

    async def serve(self, debug=False):
        "Run the scheduler"
        self.session.config.debug = debug
        self.session.set_as_default()
        await self.session.serve()

    def set_logger(self):
        warnings.warn((
            "set_logger is deprecated and will be removed in the future. "
            "Please set the logger using logging.getLogger"
        ), DeprecationWarning)
        def wrapper(func):
            func(self._get_task_logger())
        return wrapper

    def _get_task_logger(self):
        logger_name = self.session.config.task_logger_basename
        return logging.getLogger(logger_name)

    def _set_logger_with_repo(self, repo):
        if repo is None:
            repo = MemoryRepo(model=LogRecord)
        logger = self._get_task_logger()
        logger.handlers.insert(0, RepoHandler(repo=repo))

class Grouper(_AppMixin):

    def __init__(self, prefix:str=None, start_cond=None, execution=None):
        self.prefix = prefix
        self.start_cond = start_cond
        self.execution = execution

        # task_execution here should not matter
        self.session = Session(config={'task_execution': 'process'})



from pathlib import Path
import types

import pytest

from rocketry.tasks import FuncTask
from rocketry.conditions import AlwaysTrue
from rocketry.parse.utils import ParserError

def myfunc(): ...

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_construct(tmpdir, session, execution):

    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:
        if execution == "process":
            with pytest.raises(AttributeError):
                # Unpicklable function (cannot use process)
                task = FuncTask(lambda : None, execution=execution, session=session)
        else:
            task = FuncTask(lambda : None, execution=execution, session=session)

        # This should always be picklable
        task = FuncTask(myfunc, execution=execution, session=session)
        assert not task.is_delayed()
        assert task.status is None

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_construct_callable_class(tmpdir, session, execution):
    class MyClass:
        def __call__(self):
            pass

    # Going to tempdir to dump the log files there
    with tmpdir.as_cwd() as old_dir:
        # This should always be picklable
        task = FuncTask(MyClass(), execution=execution, session=session)
        assert not task.is_delayed()
        assert task.status is None
        assert task.name.endswith("test_construct:MyClass")

def test_description(session):
    "Test Task.description is correctly set (uses the func if missing)"

    # Test no description
    task = FuncTask(name="no desc 1", session=session)
    @task
    def myfunc():
        ...
    assert task.description is None

    task = FuncTask(lambda x: x, name="no desc 2", execution="main", session=session)
    assert task.description is None


     # Test description from doc (decorated)
    task = FuncTask(name="has desc 1", session=session)
    @task
    def myfunc():
        "This is func"
    assert task.description == "This is func"

    # Test description from doc (normal)
    def myfunc():
        "This is func"
    task = FuncTask(myfunc, name="has desc 2", session=session)
    assert task.description == "This is func"

    # Test the description is respected if doc is found
    def myfunc():
        "This is func"
    task = FuncTask(myfunc, name="has desc 3", description="But this is preferred", session=session)
    assert task.description == "But this is preferred"

    # Test the description is respected if doc missing
    def myfunc():
        ...
    task = FuncTask(myfunc, name="has desc 4", description="This is func", session=session)
    assert task.description == "This is func"

@pytest.mark.parametrize("execution", ["main", "thread", "process"])
def test_construct_delayed(session, execution, tmpdir):
    tmpdir.join("myfile.py").write("")
    with tmpdir.as_cwd():
        with pytest.warns(UserWarning):
            task = FuncTask(func_name="myfunc", path="missing.py", execution=execution, session=session)
        task = FuncTask(func_name="myfunc", path="myfile.py", execution=execution, session=session)
        assert task.status is None
        assert task.is_delayed()
        assert task.func_name == "myfunc"
        assert task.path == Path("myfile.py")
        assert task.func is None

def test_construct_decorate(session):
    @FuncTask(start_cond=AlwaysTrue(), name="mytask", execution="main", session=session)
    def do_stuff():
        pass

    assert isinstance(do_stuff, types.FunctionType)

    do_stuff_task = session["mytask"]
    assert isinstance(do_stuff_task, FuncTask)
    assert do_stuff_task.status is None
    assert do_stuff_task.start_cond == AlwaysTrue()
    assert do_stuff_task.name == "mytask"

    assert {do_stuff_task} == session.tasks

def test_construct_decorate_minimal(session):
    """This is an exception when FuncTask returns itself
    (__init__ cannot return anything else)"""
    # Going to tempdir to dump the log files there
    orig_default_exec = session.config.task_execution
    session.config.task_execution = 'main'
    try:
        with pytest.warns(UserWarning):
            @FuncTask
            def do_stuff():
                pass

        assert set() == session.tasks
    finally:
        session.config.task_execution = orig_default_exec

def test_construct_decorate_default_name(session):

    @FuncTask(start_cond=AlwaysTrue(), execution="main", session=session)
    def do_stuff():
        pass

    assert isinstance(do_stuff, types.FunctionType)
    do_stuff_task = list(session.tasks)[-1]
    assert isinstance(do_stuff_task, FuncTask)
    assert do_stuff_task.status is None
    assert do_stuff_task.start_cond == AlwaysTrue()
    assert do_stuff_task.name.endswith(":do_stuff")

    assert [do_stuff_task] == list(session.tasks)

@pytest.mark.parametrize(
    "start_cond,depend,expected",
    [
        pytest.param(
            AlwaysTrue(),
            None,
            AlwaysTrue(),
            id="AlwaysTrue"),
    ],
)
def test_set_start_condition(start_cond, depend, expected, session):

    task = FuncTask(
        lambda : None,
        name="task",
        start_cond=start_cond,
        execution="main",
        session=session
    )
    assert expected == task.start_cond


@pytest.mark.parametrize(
    "start_cond_str,start_cond",
    [
        pytest.param("true", lambda: AlwaysTrue(), id="true"),
        pytest.param("always true & always true", lambda: AlwaysTrue() & AlwaysTrue(), id="always true & always true"),
    ],
)
def test_set_start_condition_str(start_cond_str, start_cond, session):

    task = FuncTask(
        lambda : None,
        name="task",
        start_cond=start_cond_str,
        execution="main",
        session=session
    )
    assert start_cond() == task.start_cond

    assert str(task.start_cond) == start_cond_str

@pytest.mark.parametrize(
    "get_task,exc",
    [
        pytest.param(
            lambda s: FuncTask(lambda : None, name="task", start_cond="this is not valid", execution="main", session=s),
            ParserError,
            id="invalid start_cond"
        ),
        pytest.param(
            lambda s: FuncTask(lambda : None, name="task", execution="not valid", session=s),
            ValueError,
            id="invalid execution"
        ),
    ],
)
def test_failure(session, exc, get_task):
    with pytest.raises(exc):
        get_task(session)
    assert session.tasks == set()

def test_rename(session):
    task1 = FuncTask(lambda : None, name="a task 1", execution="main", session=session)
    task2 = FuncTask(lambda : None, name="a task 2", execution="main", session=session)
    assert session.tasks == {task1, task2}
    assert 'renamed task' not in session
    assert 'a task 1' in session

    task1.name = "renamed task"
    assert task1.name == "renamed task"
    assert session.tasks == {task1, task2}
    assert 'renamed task' in session
    assert 'a task 1' not in session

def test_rename_conflict(session):
    task1 = FuncTask(lambda : None, name="a task 1", execution="main", session=session)
    task2 = FuncTask(lambda : None, name="a task 2", execution="main", session=session)
    assert session.tasks == {task1, task2}
    assert 'renamed task' not in session
    assert 'a task 1' in session

    with pytest.raises(ValueError):
        task1.name = "a task 2"
    assert session.tasks == {task1, task2}
    assert session['a task 2'] is task2
    assert session['a task 2'] is not task1

def test_existing_raise(session):
    assert session.config.task_pre_exist == 'raise'

    task1 = FuncTask(lambda : None, name="a task", execution="main", session=session)
    with pytest.raises(ValueError):
        task2 = FuncTask(lambda : None, name="a task", execution="main", session=session)
    assert session.tasks == {task1}

def test_existing_ignore(session):
    session.config.task_pre_exist = 'ignore'
    task1 = FuncTask(lambda : None, name="a task", execution="main", session=session)
    task2 = FuncTask(lambda : None, name="a task", execution="main", session=session)
    assert session.tasks == {task1}

@pytest.mark.skip(reason="No support for this yet")
def test_existing_replace(session):
    session.config.task_pre_exist = 'replace'
    task1 = FuncTask(lambda : None, name="a task", execution="main", session=session)
    task2 = FuncTask(lambda : None, name="a task", execution="main", session=session)
    assert session.tasks == {"a task": task2}

def test_existing_rename(session):
    session.config.task_pre_exist = 'rename'
    task1 = FuncTask(lambda : None, name="a task", execution="main", session=session)
    task2 = FuncTask(lambda : None, name="a task", execution="main", session=session)
    assert session.tasks == {task1, task2}
    assert task2.name == 'a task - 1'

    assert session['a task'] is task1
    assert session['a task - 1'] is task2



from .session import Session
from .application import Rocketry, Grouper
from . import _version
from .core import Scheduler

from ._setup import _setup_defaults
from . import (
    conditions,
    log,

    args,
    time,
    tasks,
)
from .tasks import FuncTask
_setup_defaults()
session = Session(config={"task_execution": "process"})
session.set_as_default()


__version__ = _version.get_versions()['version']
